from __future__ import annotations

"""
Standalone Logging Module that doesn't depend on any other modules
"""


import re
import os
import sys
import logging
import typing
import traceback
import warnings
import pprint
import atexit as _atexit
import functools
import threading
from enum import Enum

from loguru import _defaults
from loguru._logger import Core as _Core
from loguru._logger import Logger as _Logger
from typing import Type, Union, Optional, Any, List, Dict, Tuple, Callable, Set, TYPE_CHECKING

# Use this section to filter out warnings from other modules
os.environ['TF_CPP_MIN_LOG_LEVEL'] = os.getenv('TF_CPP_MIN_LOG_LEVEL', '3')
warnings.filterwarnings('ignore', message='Unverified HTTPS request')
warnings.filterwarnings("ignore", message = "Unclosed client session")

if TYPE_CHECKING:
    from lazyops.types.models import BaseSettings




DEFAULT_STATUS_COLORS = {
    'debug': '<fg #D9ED92>',
    'info': '<fg #34A0A4>',
    'success': '<fg #52B69A>',
    'warning': '<fg #F48C06>',
    'error': '<fg #DC2F02>',
    'critical': '<fg #9D0208>',
}
    

QUEUE_STATUS_COLORS = {
    'new': '<fg #D9ED92>',
    'deferred': '<fg #B5E48C>',
    'queued': '<fg #99D98C>',
    'active': '<fg #76C893>',
    'complete': '<fg #52B69A>',

    # Error Colors
    'aborted': '<fg #FFBA08>',
    'failed': '<fg #9D0208>',

    # Other Colors
    'enqueue': '<fg #168AAD>',
    'finish': '<fg #52B69A>',
    'completed': '<fg #52B69A>',
    'error': '<fg #DC2F02>',
    'abort': '<fg #DC2F02>',

    'retry': '<fg #F48C06>',
    'scheduled': '<fg #34A0A4>',
    'startup': '<fg #168AAD>',
    'shutdown': '<fg #6A040F>',
    'process': '<fg #184E77>',
    'sweep': '<fg #B5E48C>',
    'stats': '<fg #B5E48C>',
    'dequeue': '<fg #168AAD>',
}

STATUS_COLOR = QUEUE_STATUS_COLORS

FALLBACK_STATUS_COLOR = '<fg #99D98C>'

# DEFAULT_FUNCTION_COLOR = '<fg #457b9d>'
DEFAULT_FUNCTION_COLOR = '<fg #219ebc>'
DEFAULT_CLASS_COLOR = '<fg #a8dadc>'

RESET_COLOR = '\x1b[0m'

LOGLEVEL_MAPPING = {
    50: 'CRITICAL',
    40: 'ERROR',
    30: 'WARNING',
    20: 'INFO',
    19: 'DEV',
    10: 'DEBUG',
    5: 'CRITICAL',
    4: 'ERROR',
    3: 'WARNING',
    2: 'INFO',
    1: 'DEBUG',
    0: 'NOTSET',
}
REVERSE_LOGLEVEL_MAPPING = {v: k for k, v in LOGLEVEL_MAPPING.items()}

COLORED_MESSAGE_MAP = {
    '|bld|': '<bold>',
    '|reset|': '</>',
    '|eee|': '</></></>',
    '|em|': '<bold>',
    '|ee|': '</></>',
    '|lr|': '<light-red>',
    '|lb|': '<light-blue>',
    '|lm|': '<light-magenta>',
    '|lc|': '<light-cyan>',
    '|lw|': '<light-white>',
    '|gr|': '<gray>',
    '|lk|': '<light-black>',
    '|br|': "\x1b[31;1m", # Bold Red
    '|k|': '<black>',
    '|r|': '<red>',
    '|m|': '<magenta>',
    '|c|': '<cyan>',
    '|u|': '<underline>',
    '|i|': '<italic>',
    '|s|': '<strike>',
    '|e|': '</>',
    '|g|': '<green>',
    '|y|': '<yellow>',
    '|b|': '<blue>',
    '|w|': '<white>',
}


def find_and_format_seps(msg: str) -> str:
    """
    Find any |a,b,c| and format them |a||b||c|

    ex:
      |em,b,u| -> |em||b||u|
      |em,b| -> |em||b|
      
    """
    # v2
    for sep_match in re.finditer('\|\w+,(\w+,*)+\|', msg):
        s = sep_match.group()
        if len(s) >= 10: continue
        msg = msg.replace(s, "||".join(s.split(",")))
    return msg

def extract_module_name(name: str) -> str:
    """
    Extracts the module name from a name
    """
    return name.split('.')[0]

class NullLogger(logging.Logger):
    """
    A logger that does nothing
    """
    def info(self, *args, **kwargs): pass
    def debug(self, *args, **kwargs): pass
    def warning(self, *args, **kwargs): pass
    def error(self, *args, **kwargs): pass
    def critical(self, *args, **kwargs): pass
    def exception(self, *args, **kwargs): pass
    def log(self, *args, **kwargs): pass
    def trace(self, *args, **kwargs): pass
    def success(self, *args, **kwargs): pass


# Setup Default Logger
class Logger(_Logger):

    name: str = None
    settings: Type['BaseSettings'] = None
    conditions: Dict[str, Tuple[Union[Callable, bool], str]] = {}
    default_trace_depth: Optional[int] = None
    is_global: bool = False

    _colored_opts = None
    _current_level: Optional[str] = None


    @property
    def colored_opts(self):
        """
        Returns the colored options
        """
        if not self._colored_opts:
            (exception, depth, record, lazy, colors, raw, capture, patchers, extra) = self._options
            self._colored_opts = (exception, depth, record, lazy, True, raw, capture, patchers, extra)
        return self._colored_opts

    def _get_opts(self, colored: Optional[bool] = False, **kwargs):
        """
        Returns the options
        """
        return self.colored_opts if colored else self._options


    def get_log_mode(self, level: str = "info"):
        """
        Returns the log mode based on the level
        """
        return self.dev if level.upper() in {'DEV'} else getattr(self, level.lower())
    

    def add_if_condition(
        self, 
        name: str, 
        condition: Union[Callable, bool],
        level: Optional[Union[str, int]] = 'INFO',
    ):
        """
        Adds a condition to the logger
        """
        self.conditions[name] = (condition, self._get_level(level))
    
    def remove_if_condition(self, name: str):
        """
        Removes a condition from the logger
        """
        if name in self.conditions:
            del self.conditions[name]

    def _is_dev_condition(self, record: logging.LogRecord) -> bool:
        """
        Returns whether the dev condition is met
        """
        if not self.settings: return True
        if record.levelname == 'DEV':
            for key in {'api_dev_mode', 'debug_enabled'}:
                if (
                    hasattr(self.settings, key)
                    and getattr(self.settings, key) is False
                ):
                    return False
        return True

    def _filter_if(self, name: str, record: Optional[logging.LogRecord] = None, message: Optional[Any] = None, level: Optional[Union[str, int]] = None) -> Tuple[bool, str]:
        """
        Filters out messages based on conditions
        """
        if name in self.conditions:
            condition, clevel = self.conditions[name]
            if isinstance(condition, bool):
                return condition, clevel
            elif isinstance(condition, type(None)):
                return False, clevel
            elif isinstance(condition, Callable):
                return condition(record or message), clevel
        return True, (record.levelname if record else self._get_level(level or 'INFO'))
    
    def _filter_module_name(self, name: str) -> bool:
        """
        Filter based on module name

        - True if the module is not registered and is_global is False 
        - False if the module is registered and is_global is False
        """
        _is_registered = is_registered_logger_module(name)
        if self.is_global: 
            return _is_registered is not False
        return _is_registered is False


    def _filter(self, record: logging.LogRecord, name: Optional[str] = None) -> bool:
        """
        Filters out messages based on conditions

        - True if the message should be filtered out
        - False if the message should be logged
        """
        if self._filter_module_name(record['name']): 
            return True
        if name is not None:
            return self._filter_if(name, record)[0]
        if not self.conditions: return False
        return not any(
            isinstance(value, bool)
            and value is False
            or not isinstance(value, bool)
            and isinstance(value, Callable)
            and value(record) is False
            for key, value in self.conditions.items()
        )

    def _filter_dev(self, record: logging.LogRecord, **kwargs):
        if not self.settings:
            return True
        if record.levelname == 'DEV':
            for key in {'api_dev_mode', 'debug_enabled'}:
                if (
                    hasattr(self.settings, key)
                    and getattr(self.settings, key) is False
                ):
                    return False
        return True
    
    def opt(
        self,
        *,
        exception=None,
        record=False,
        lazy=False,
        colors=False,
        raw=False,
        capture=True,
        depth=0,
        ansi=False
    ):
        """
        Return a new logger with the specified options changed.
        """
        if ansi: colors = True
        args = self._options[-2:]
        return type(self)(self._core, exception, depth, record, lazy, colors, raw, capture, *args)



    """
    Newly Added APIs
    """

    def _get_level(self, level: Union[str, int]) -> str:
        """
        Returns the log level
        """
        if isinstance(level, str): level = level.upper()
        elif isinstance(level, int): level = LOGLEVEL_MAPPING.get(level, 'INFO')
        return level

    def _format_item(
        self,
        msg: Any,
        max_length: Optional[int] = None,
        colored: Optional[bool] = False,
        level: Optional[str] = None,
        _is_part: Optional[bool] = False,
    ) -> str:  # sourcery skip: extract-duplicate-method, low-code-quality, split-or-ifs
        """
        Formats an item
        """
        # Primitive Types
        if isinstance(msg, str): return msg[:max_length] if max_length else msg
        if isinstance(msg, (float, int, bool, type(None))): return str(msg)[:max_length] if max_length else str(msg)
        if isinstance(msg, (list, set)):
            _msg = str(msg) if _is_part else "".join(f'- {item}\n' for item in msg)
            return _msg[:max_length] if max_length else _msg
        
        prefix, suffix = '', ''
        if colored:
            prefix = '|g|'
            if level:
                level = level.lower()
                prefix = DEFAULT_STATUS_COLORS.get(level, '|g|')
            suffix = '|e|'

        if isinstance(msg, dict):
            _msg = "\n"
            for key, value in msg.items():
                _value = f'{value}'
                if max_length and len(_value) > max_length:
                    _value = f'{_value[:max_length]}...'
                _msg += f'- {prefix}{key}{suffix}: {_value}\n'
            return _msg.rstrip()
        
        if isinstance(msg, tuple):
            _msg = "".join(f'- {prefix}{key}{suffix}: {self._format_item(value, max_length = max_length, colored = colored, level = level,  _is_part = True)}\n' for key, value in zip(msg[0], msg[1]))
            return _msg[:max_length] if max_length else _msg

        # Complex Types
        if (hasattr(msg, 'dict') and hasattr(msg, 'Config')) or hasattr(msg, 'fields'):
            # Likely Pydantic Model
            _msg = f'{prefix}[{msg.__class__.__name__}]{suffix}'
            fields = msg.fields.keys() if hasattr(msg, 'fields') else msg.__fields__.keys()
            for field in fields:
                field_str = f'{prefix}{field}{suffix}'
                val_s = f'\n  {field_str}: {getattr(msg, field)!r}'
                if max_length is not None and len(val_s) > max_length:
                    val_s = f'{val_s[:max_length]}...'
                _msg += val_s
            return _msg
            # return self._format_item(msg.dict(), max_length = max_length, colored = colored, _is_part = _is_part)
        
        if hasattr(msg, 'model_dump'):
            return self._format_item(msg.model_dump(mode = 'json'), max_length = max_length, colored = colored, level = level, _is_part = _is_part)

        if hasattr(msg, 'dict'):
            return self._format_item(msg.dict(), max_length = max_length, colored = colored, level = level, _is_part = _is_part)

        if hasattr(msg, 'json'):
            return self._format_item(msg.json(), max_length = max_length, colored = colored, level = level, _is_part = _is_part)

        if hasattr(msg, '__dict__'):
            return self._format_item(msg.__dict__, max_length = max_length, colored = colored, level = level, _is_part = _is_part)
        
        return str(msg)[:max_length] if max_length else str(msg)


    def _format_message(
        self, 
        message: Any, 
        *args,
        prefix: Optional[str] = None,
        max_length: Optional[int] = None,
        level: Optional[str] = None,
        colored: Optional[bool] = False,
    ) -> str:
        """
        Formats the message

        "example |b|msg|e|"
        -> "example <blue>msg</><reset>"
        """
        _message = ""
        if prefix: 
            if colored and '|' not in prefix: 
                base_color = '|g|'
                if level:
                    level = level.lower()
                    base_color = DEFAULT_STATUS_COLORS.get(level, '|g|')
                prefix = f'{base_color}{prefix}|e|'
            _message += f'[{prefix}] '
        _message += self._format_item(message, max_length = max_length, colored = colored, level = level)
        if args:
            for arg in args:
                _message += "\n"
                _message += self._format_item(arg, max_length = max_length, colored = colored, level = level)
        if colored:
            # Add escape characters to prevent errors
            _message = _message.replace("<fg", ">|fg")
            _message = _message.replace("<", "\<")
            _message = find_and_format_seps(_message)
            # print(_message)
            for key, value in COLORED_MESSAGE_MAP.items():
                _message = _message.replace(key, value)
            _message = _message.replace(">|fg", "<fg")
            _message += RESET_COLOR
        return _message

    
    def log_if(
        self, 
        name: str, 
        message: Any, 
        *args, 
        level: Optional[Union[str, int]] = None, 
        **kwargs
    ):  # noqa: N805
        """
        Log ``message.format(*args, **kwargs)`` with severity ``level`` if condition is met.
        """
        condition, clevel = self._filter_if(name, message = message, level = level)
        if condition:
            return self.log((level or clevel), message, *args, **kwargs)

    def log(
        self, 
        level: Union[str, int], 
        message: Any, 
        *args, 
        prefix: Optional[str] = None,
        max_length: Optional[int] = None,
        colored: Optional[bool] = False,
        **kwargs
    ):  # noqa: N805
        """
        Log ``message.format(*args, **kwargs)`` with severity ``level``.
        """
        level = self._get_level(level)
        message = self._format_message(message, prefix = prefix, max_length = max_length, colored = colored, level = level)
        try:
            self._log(level, False, self._get_opts(colored = colored), message, args, kwargs)
        except TypeError:
            # Compatibility with < 0.6.0
            # level_id, static_level_no, from_decorator, options, message, args, kwargs
            static_log_no = REVERSE_LOGLEVEL_MAPPING.get(level, 20)
            self._log(level, static_log_no, False, self._get_opts(colored = colored), message, args, kwargs)

    def info(
        self, 
        message: Any, 
        *args, 
        colored: Optional[bool] = False, 
        prefix: Optional[str] = None,
        max_length: Optional[int] = None,
        **kwargs
    ):  # noqa: N805
        """
        Log ``message.format(*args, **kwargs)`` with severity ``'INFO'``.
        """
        message = self._format_message(message, *args, prefix = prefix, max_length = max_length, colored = colored, level = 'INFO')
        try:
            self._log("INFO", False, self._get_opts(colored = colored), message, args, kwargs)
        except TypeError:
            # Compatibility with < 0.6.0
            self._log("INFO", 20, False, self._get_opts(colored = colored), message, args, kwargs)

    def success(
        self, 
        message, 
        *args, 
        colored: Optional[bool] = False, 
        prefix: Optional[str] = None,
        max_length: Optional[int] = None,
        **kwargs
    ):  # noqa: N805
        r"""Log ``message.format(*args, **kwargs)`` with severity ``'SUCCESS'``."""
        message = self._format_message(message, *args, prefix = prefix, max_length = max_length, colored = colored, level = 'SUCCESS')
        try:
            self._log("SUCCESS", False, self._get_opts(colored = colored), message, args, kwargs)
        except TypeError:
            # Compatibility with < 0.6.0
            self._log("SUCCESS", 20, False, self._get_opts(colored = colored), message, args, kwargs)
    

    def warning(
        self, 
        message, 
        *args, 
        colored: Optional[bool] = False, 
        prefix: Optional[str] = None,
        max_length: Optional[int] = None,
        **kwargs
    ):  # noqa: N805
        r"""Log ``message.format(*args, **kwargs)`` with severity ``'WARNING'``."""
        message = self._format_message(message, prefix = prefix, max_length = max_length, colored = colored, level = 'WARNING')
        try:
            self._log("WARNING", False, self._get_opts(colored = colored), message, args, kwargs)
        except TypeError:
            # Compatibility with < 0.6.0
            self._log("WARNING", 30, False, self._get_opts(colored = colored), message, args, kwargs)


    def trace(
        self, 
        msg: Union[str, Any], 
        error: Optional[Type[Exception]] = None, 
        level: str = "ERROR",
        limit: Optional[int] = None,
        chain: Optional[bool] = True,
        colored: Optional[bool] = False,
        prefix: Optional[str] = None,
        max_length: Optional[int] = None,
        **kwargs,
    ) -> None:
        """
        This method logs the traceback of an exception.

        :param error: The exception to log.
        """
        _depth = kwargs.pop('depth', None)
        if _depth is not None: limit = _depth
        _msg = msg if isinstance(msg, str) else self._format_message(msg, colored = colored, level = level, prefix = prefix, max_length = max_length)
        # pprint.pformat(msg)
        _msg += f"\n{traceback.format_exc(chain = chain, limit = limit)}"
        if error: _msg += f" - {error}"
        try:
            self._log(level, False, self._get_opts(colored = colored), _msg, (), {})
        except TypeError:
            static_log_no = REVERSE_LOGLEVEL_MAPPING.get(level, 40)
            self._log(level, static_log_no, False, self._get_opts(colored = colored), _msg, (), {})

    def exception(
        self,
        message: Any,
        *args,
        colored: Optional[bool] = False,
        prefix: Optional[str] = None,
        max_length: Optional[int] = None,
        **kwargs
    ):
        """
        Log ``message.format(*args, **kwargs)`` with severity ``'ERROR'``.
        """
        message = self._format_message(message, prefix = prefix, max_length = max_length, colored = colored, level = 'ERROR')
        super().exception(message, *args, **kwargs)

    
    def __call__(self, message: Any, *args, level: str = 'info', **kwargs):
        r"""Log ``message.format(*args, **kwargs)`` with severity ``'INFO'``."""
        if isinstance(message, list):
            __message = "".join(f'- {item}\n' for item in message)
        elif isinstance(message, dict):
            __message = "".join(f'- {key}: {value}\n' for key, value in message.items())
        else:
            __message = str(message)
        _log = self.get_log_mode(level)
        _log(__message.strip(), *args, **kwargs)
    
    def _logcompat(
        self, level, from_decorator, options, message, args, kwargs
    ):
        """
        Compatible to < 0.6.0
        """
        try:
            self._log(level, from_decorator, options, message, args, kwargs)
        except TypeError:
            # Compatibility with < 0.6.0
            # level_id, static_level_no, from_decorator, options, message, args, kwargs
            static_log_no = REVERSE_LOGLEVEL_MAPPING.get(level, 20)
            self._log(level, static_log_no, from_decorator, options, message, args, kwargs)


    """
    Utilties
    """
    def render(
        self,
        objs: Union[Any, List[Any]],
        level: str = 'INFO',
        **kwargs,
    ) -> None:
        """
        Tries to render an object with pformat
        """
        if not isinstance(objs, list):
            objs = [objs]
        _log = self.get_log_mode(level)
        for obj in objs:
            try:
                _log('\n' + pprint.pformat(obj, **kwargs))
            except Exception as e:
                _log('\n' + obj)

    def render_yaml(
        self,
        objs: Union[Any, List[Any]],
        level: str = 'INFO',
        **kwargs,
    ) -> None:
        """
        Tries to render a yaml object.
        """
        _log = self.get_log_mode(level)
        try:
            from fileio.io import Yaml
            _log('\n' + Yaml.dumps(objs, **kwargs))
        
        except Exception as e:
            import yaml
            _log('\n' + yaml.dump(objs, **kwargs))

    def display_crd(self, message: Any, *args, level: str = 'info', **kwargs):
        """
        Display CRD information in the log.

        Parameters
        ----------
        message : Any
            The message to display.
        level : str
            The log level to use.
        """
        __message = ""
        if isinstance(message, list):
            for m in message:
                if isinstance(m, dict):
                    __message += "".join(f'- <light-blue>{key}</>: {value}\n' for key, value in m.items())
                else:
                    __message += f'- <light-blue>{m}</>\n'
        elif isinstance(message, dict):
            __message = "".join(f'- <light-blue>{key}</>: {value}\n' for key, value in message.items())
        else:
            __message = str(message)
        _log = self.get_log_mode(level)
        _log(__message.strip(), *args, **kwargs)


    def mute_logger(self, modules: Optional[Union[str, List[str]]], level: str = 'WARNING'):
        """
        Helper to mute a logger from another module.
        """
        if not isinstance(modules, list):
            modules = [modules]
        for module in modules:
            logging.getLogger(module).setLevel(logging.getLevelName(level))

    def mute_warning(self, action: str = 'ignore', category: Type[Warning] = Warning, module: str = None, **kwargs):
        """
        Helper to mute a warning from another module.
        """
        if module:
            warnings.filterwarnings(action, category=category, module=module, **kwargs)
        else:
            warnings.filterwarnings(action, category=category, **kwargs)
    
    def change_logger_level(
        self,
        level: str,
    ):
        """
        Changes the logger level

        :TODO
        """
        return
        # self._core.configure(levels = [level])

    
    def add_healthz_filter(
        self,
        loggers: Optional[Union[str, List[str]]] = None,
        paths: Optional[Union[str, List[str]]] = None,
    ):
        """
        Adds a filter to the logger to filter out healthz requests.

        Parameters
        ----------
        loggers : Optional[Union[str, List[str]]]
            The loggers to add the filter to.
            defaults to ['gunicorn.glogging.Logger', 'uvicorn.access']
        paths : Optional[Union[str, List[str]]]
            The paths to filter out.
            defaults to ['/healthz', '/health']
        """
        if not loggers: loggers = ['gunicorn.glogging.Logger', 'uvicorn.access']
        if not paths: paths = ['/healthz', '/health']
        if not isinstance(loggers, list): loggers = [loggers]
        if not isinstance(paths, list): paths = [paths]

        def _healthz_filter(record: logging.LogRecord) -> bool:
            if 'path' in record.args:
                # if record.args.get('path'):
                return record.args['path'] not in paths
            return all(path not in record.args for path in paths)

        for logger in loggers:
            logging.getLogger(logger).addFilter(_healthz_filter)

class InterceptHandler(logging.Handler):
    loglevel_mapping = LOGLEVEL_MAPPING

    def emit(self, record):
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = self.loglevel_mapping.get(record.levelno, 'DEBUG')
        # if "Unclosed client session" in record.message:
        #     print('Has unclosed client session')
        #     return
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1
        log = logger.bind(request_id=None)
        log.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())



class LoggerFormatter:

    max_extra_lengths: Dict[str, int] = {}

    @classmethod
    def get_extra_length(cls, key: str, value: str) -> int:
        """
        Returns the max length of an extra key
        """
        if key not in cls.max_extra_lengths:
            cls.max_extra_lengths[key] = len(key)
        if len(value) > cls.max_extra_lengths[key]:
            cls.max_extra_lengths[key] = len(value)
        return cls.max_extra_lengths[key]

    @classmethod
    def queue_logger_formatter(cls, record: Dict[str, Union[Dict[str, Any], Any]]) -> str:
        """
        Formats the log message for the queue.
        """
        _extra: Dict[str, Union[Dict[str, Any], Any]] = record.get('extra', {})
        if not record['extra'].get('worker_name'):
            record['extra']['worker_name'] = ''
        
        status = _extra.get('status')
        kind: str = _extra.get('kind')
        if status and isinstance(status, Enum): status = status.name
        
        kind_color = QUEUE_STATUS_COLORS.get(kind.lower(), FALLBACK_STATUS_COLOR)
        if '<' not in kind_color: kind_color = f'<{kind_color}>'
        extra = kind_color + '{extra[kind]}</>:'
        if _extra.get('queue_name'):
            queue_name_length = cls.get_extra_length('queue_name', _extra['queue_name'])
            extra += '<b><fg #006d77>{extra[queue_name]:<' + str(queue_name_length) + '}</></>:'
        if _extra.get('worker_name'):
            worker_name_length = cls.get_extra_length('worker_name', _extra['worker_name'])
            extra += '<fg #83c5be>{extra[worker_name]:<' + str(worker_name_length) + '}</>:'
        # extra += '<fg #83c5be>{extra[worker_name]}</>:<b><fg #006d77>{extra[queue_name]:<18}</></>:'
        if _extra.get('job_id'):
            extra += '<fg #005f73>{extra[job_id]}</>'
        if status:
            status_color = QUEUE_STATUS_COLORS.get(status.lower(), FALLBACK_STATUS_COLOR)
            if '<' not in status_color: status_color = f'<{status_color}>'
            extra += f':{status_color}' + '{extra[status]}</>: '
        # extra += RESET_COLOR
        # print(extra)
        return extra


    @classmethod
    def default_formatter(cls, record: Dict[str, Union[Dict[str, Any], Any]]) -> str:
        """
        To add a custom format for a module, add another `elif` clause with code to determine `extra` and `level`.

        From that module and all submodules, call logger with `logger.bind(foo='bar').info(msg)`.
        Then you can access it with `record['extra'].get('foo')`.
        """        
        _extra = record.get('extra', {})
        extra = DEFAULT_CLASS_COLOR + '{name}</>:' + DEFAULT_FUNCTION_COLOR + '{function}</>: '
        if _extra.get('worker_name') or _extra.get('queue_name'):
            extra = cls.queue_logger_formatter(record)
        
        if 'result=tensor([' not in str(record['message']):
            return "<level>{level: <8}</> <green>{time:YYYY-MM-DD HH:mm:ss.SSS}</>: " \
                       + extra + "<level>{message}</level>" + RESET_COLOR + "\n"
        
        msg = str(record['message'])[:100].replace('{', '(').replace('}', ')')
        return "<level>{level: <8}</> <green>{time:YYYY-MM-DD HH:mm:ss.SSS}</>: "\
                   + extra + "<level>" + msg + f"</level>{RESET_COLOR}\n"



_registered_logger_modules: Set[str] = set()

def register_logger_module(module: str):
    """
    Registers a logger module
    """
    global _registered_logger_modules
    module = extract_module_name(module)
    _registered_logger_modules.add(module)
    # print('Registered logger module: ', module)


@functools.lru_cache(maxsize=1000)
def is_registered_logger_module(name: str) -> bool:
    """
    Returns whether a logger module is registered
    """
    module_name = extract_module_name(name)
    return module_name in _registered_logger_modules



_lock = threading.Lock()
_logger_contexts: Dict[str, Logger] = {
    # 'lazyops': _logger,
}

def create_global_logger(
    name: Optional[str] = 'lazyops',
    level: Union[str, int] = "INFO",
    format: Optional[Callable] = None,
    filter: Optional[Callable] = None,
    handlers: Optional[List[logging.Handler]] = None,
    settings: Optional['BaseSettings'] = None,
    **kwargs,
) -> Logger:
    """
    Creates the global logger
    """
    global _logger_contexts

    # < 0.7.0
    try:
        _logger = Logger(
            core=_Core(),
            exception=None,
            depth=0,
            record=False,
            lazy=False,
            colors=True,
            raw=False,
            capture=True,
            patcher=None,
            extra={},
        )
    # >= 0.7.0
    except Exception as e:
        _logger = Logger(
            core=_Core(),
            exception=None,
            depth=0,
            record=False,
            lazy=False,
            colors=False,
            raw=False,
            capture=True,
            patchers=[],
            extra={},
        )

    _logger.name = 'lazyops'
    _logger.is_global = True
    dev_level = _logger.level(name='DEV', no=19, color="<blue>", icon="@")

    if _defaults.LOGURU_AUTOINIT and sys.stderr: _logger.add(sys.stderr)
    _atexit.register(_logger.remove)

    _logger.remove()
    _logger.add_if_condition('dev', _logger._is_dev_condition)
    
    logging.basicConfig(
        handlers = handlers or [InterceptHandler()],  level = 0
    )

    _logger.add(
        sys.stdout,
        enqueue = True,
        backtrace = True,
        colorize = True,
        level = level,
        format = format if format is not None else LoggerFormatter.default_formatter,
        filter = filter if filter is not None else _logger._filter,
        **kwargs,
    )
    if settings: _logger.settings = settings
    _logger_contexts[name] = _logger
    return _logger


def create_default_logger(
    name: Optional[str] = None,
    level: Union[str, int] = "INFO",
    format: Optional[Callable] = None,
    filter: Optional[Callable] = None,
    handlers: Optional[List[logging.Handler]] = None,
    settings: Optional['BaseSettings'] = None,
    **kwargs,
) -> Logger:
    """
    Creates a default logger
    """
    global _logger_contexts

    if name:
        if name.upper() in REVERSE_LOGLEVEL_MAPPING:
            # If name is a level, then set level to name
            level = name
            name = None
        else:
            name = extract_module_name(name)
            
    if name is None: name = 'lazyops'
    if name in _logger_contexts:
        return _logger_contexts[name]
    
    with _lock:
        if name == 'lazyops':
            return create_global_logger(
                name = name,
                level = level,
                format = format,
                filter = filter,
                handlers = handlers,
                settings = settings,
            )
        
        if isinstance(level, str): level = level.upper()
        # print(f"[{name}] Creating logger with level: {level}")
        _logger = _logger_contexts['lazyops']

        if name and format is not None:
            # Add a new handler
            def _filter_func(record: logging.LogRecord) -> bool:
                """
                Filter out messages from other modules
                """
                return extract_module_name(record.name) == name
            
            _logger.add(
                sys.stdout,
                enqueue = True,
                backtrace = True,
                colorize = True,
                level = level,
                format = format,
                filter = _filter_func,
                **kwargs,
            )
            return _logger

        *options, extra = _logger._options
        new_logger = Logger(_logger._core, *options, {**extra})
        if name: 
            _logger_contexts[name] = new_logger
            new_logger.name = name
            register_logger_module(name)

        if settings: new_logger.settings = settings
        # if _is_global_logger: new_logger.is_global = True
        return new_logger



if os.getenv('DEBUG_ENABLED') == 'True':
    logger_level = 'DEV'
else:
    logger_level: str = os.getenv('LOGGER_LEVEL', 'INFO').upper()



get_logger = create_default_logger
logger = create_default_logger(__name__, level = logger_level)

default_logger = logger
null_logger = NullLogger(name = 'null_logger')


def change_logger_level(
    name: Optional[str] = None,
    level: Union[str, int] = 'INFO',
    verbose: bool = False,
    **kwargs,
):
    """
    Change the logger level for a specific logger

    args:
        level: str = 'INFO'
            The level to change the logger to
        verbose: bool = False
            Whether to print the change to the logger
    """
    global logger, logger_level
    if isinstance(level, str): level = level.upper()
    # Skip if the level is the same
    if level == logger_level: return
    name = name or 'lazyops'
    name = name.split('.')[0]
    logger_level = level
    if name != 'lazyops':
        __logger = get_logger(name, logger_level, **kwargs)
    else:
        __logger = logger
    if verbose: __logger.info(f"[{name}] Changing logger level from {logger_level} -> {level}")
    __logger._core.min_level = float(REVERSE_LOGLEVEL_MAPPING[logger_level.upper()])
    

def add_api_log_filters(
    modules: typing.Optional[typing.Union[typing.List[str], str]] = ['gunicorn', 'uvicorn'],
    routes: typing.Optional[typing.Union[typing.List[str], str]] = ['/healthz'],
    status_codes: typing.Optional[typing.Union[typing.List[int], int]] = None,
    verbose: bool = False,
):  # sourcery skip: default-mutable-arg
    """
    Add filters to the logger to remove health checks and other unwanted logs

    args:

        modules: list of modules to filter [default: ['gunicorn', 'uvicorn']
        routes: list of routes to filter [default: ['/healthz']]
        status_codes: list of status codes to filter [default: None]
        verbose: bool = False [default: False]
    """

    if not isinstance(modules, list): modules = [modules]
    if routes and not isinstance(routes, list): routes = [routes]
    if status_codes and not isinstance(status_codes, list): status_codes = [status_codes]

    def filter_api_record(record: logging.LogRecord) -> bool:
        """
        Filter out health checks and other unwanted logs
        """
        if routes:
            for route in routes:
                if route in record.args: return False
        if status_codes:
            for sc in status_codes:
                if sc in record.args: return False
        return True
    
    for module in modules:
        if module == 'gunicorn': module = 'gunicorn.glogging.Logger'
        elif module == 'uvicorn': module = 'uvicorn.logging.Logger'
        _apilogger = logging.getLogger(module)
        if verbose: default_logger.info(f"Adding API filters to {module} for routes: {routes} and status_codes: {status_codes}")
        _apilogger.addFilter(filter_api_record)




from lazyops.types.common import UpperStrEnum
from typing import Union

UserPrivilageLevel = {
    'ANON': 0,

    # Access Level (1-20)
    'READ': 5,
    'WRITE': 10,
    'MODIFY': 15,
    'PRIVILAGED': 20, # 'PRIVILAGED' is a user with 'READ', 'WRITE', and 'MODIFY' privilages
    
    # User Level (30-50)
    'USER': 30,             # 'USER' has 'READ', 'WRITE', and 'MODIFY' privilages
    'USER_API': 30,         # 'USER_API' is a user with 'USER' and 'API' privilages
    'USER_STAFF': 35,       # 'USER_STAFF' is a user with 'USER' and 'STAFF' privilages
    'USER_PRIVILAGED': 40,  # 'USER_PRIVILAGED' is a user with 'USER' and 'PRIVILAGED' privilages
    'USER_ADMIN': 45,       # 'USER_ADMIN' is a user with 'USER' and 'ADMIN' privilages
    
    # Internal Level (50-70)
    'STAFF': 50,            # 'STAFF' has 'READ', 'WRITE', and 'MODIFY' privilages
    'API_CLIENT': 50,       # 'API_CLIENT' is a user with 'API' privilages
    'SERVICE': 50,          # 'SERVICE' is a user with 'API' privilages
    'SYSTEM': 60,           # 'SYSTEM' has 'READ', 'WRITE', and 'MODIFY' privilages
    
    # Admin Level (100+)
    'ADMIN': 100,           # 'ADMIN' has 'READ', 'WRITE', and 'MODIFY' privilages
    'SYSTEM_ADMIN': 150,
    'SUPER_ADMIN': 1000,
}
    
class UserRole(UpperStrEnum):
    ANON = 'ANON'
    
    # Access Level (1-20)
    READ = 'READ'
    WRITE = 'WRITE'
    MODIFY = 'MODIFY'
    PRIVILAGED = 'PRIVILAGED'
    
    # User Level (30-50)
    USER = 'USER'
    USER_API = 'USER_API'
    USER_STAFF = 'USER_STAFF'
    USER_PRIVILAGED = 'USER_PRIVILAGED'
    USER_ADMIN = 'USER_ADMIN'

    # Internal Level (50-70)
    STAFF = 'STAFF'
    SYSTEM = 'SYSTEM'
    SERVICE = 'SERVICE'
    API_CLIENT = 'API_CLIENT'

    # Admin Level (100+)
    ADMIN = 'ADMIN'
    SYSTEM_ADMIN = 'SYSTEM_ADMIN'
    SUPER_ADMIN = 'SUPER_ADMIN'

    @property
    def privilage_level(self) -> int:
        """
        Returns the privilage level of the user role
        - Can be subclassed to return a different value
        """
        return UserPrivilageLevel[self.value]

    def __lt__(self, other: Union[int, 'UserRole']) -> bool:
        """
        Returns True if the user role is less than the other
        """
        if isinstance(other, int):
            return self.privilage_level < other
        return self.privilage_level < other.privilage_level

    def __le__(self, other: Union[int, 'UserRole']) -> bool:
        """
        Returns True if the user role is less than or equal to the other
        """
        if isinstance(other, int):
            return self.privilage_level <= other
        return self.privilage_level <= other.privilage_level

    def __gt__(self, other: Union[int, 'UserRole']) -> bool:
        """
        Returns True if the user role is greater than the other
        """
        if isinstance(other, int):
            return self.privilage_level > other
        return self.privilage_level > other.privilage_level

    def __ge__(self, other: Union[int, 'UserRole']) -> bool:
        """
        Returns True if the user role is greater than or equal to the other
        """
        if isinstance(other, int):
            return self.privilage_level >= other
        return self.privilage_level >= other.privilage_level

    



"""
Base Import Handler
"""
import typing
import functools
import pathlib
import importlib.util
from lazyops.utils.logs import default_logger as logger
from lazyops.utils.lazylib import LazyLib
from lazyops.utils.helpers import is_coro_func

def resolve_missing(
    modules: typing.Union[str, typing.List],
    packages: typing.Union[str, typing.List] = None,
    required: bool = True,
):
    if not isinstance(modules, list):
        modules = [modules]
    if packages is not None and not isinstance(packages, list):
        packages = [packages]
    elif packages is None:
        packages = modules
    kind = 'required' if required else 'optionally required'
    logger.info(f"{', '.join(modules)} are {kind}. Installing...")
    for module, pkg in zip(modules, packages):
        LazyLib.import_lib(module, pkg)


def resolve_missing_custom(
    modules: typing.Union[str, typing.List],
    packages: typing.Union[str, typing.List] = None,
    required: bool = True,
):
    """
    Handles custom use cases like `torch` where we need to
    have a extra index to install from
    """
    if not isinstance(modules, list):
        modules = [modules]
    if packages is not None and not isinstance(packages, list):
        packages = [packages]
    elif packages is None:
        packages = modules
    
    module_names = [module.split(' ', 1)[0] for module in modules]
    kind = 'required' if required else 'optionally required'
    logger.info(f"{', '.join(module_names)} are {kind}. Installing...")
    for module, pkg in zip(modules, packages):
        module_name = LazyLib.get_requirement(module, True)
        if LazyLib.is_available(module_name):
            continue
        LazyLib.install_library(pkg)

    
def require_missing_wrapper(
    resolver: typing.Callable,
    func: typing.Callable,
    **resolver_kwargs,
):
    """
    Helper function to wrap the resolve async or sync funcs
    """
    if is_coro_func(func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            resolver(**resolver_kwargs)
            return await func(*args, **kwargs)
    else:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            resolver(**resolver_kwargs)
            return func(*args, **kwargs)

    return wrapper




from __future__ import annotations


# https://stackoverflow.com/questions/69483534/sqlalchemy-with-multiple-session-how-to-avoid-object-is-already-attached-to-se


import weakref
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.asyncio import AsyncSession, async_object_session


class CarefulSession(orm.Session):

    def add(self, object_):
        # Get the session that contains the object, if there is one.
        object_session = orm.object_session(object_)
        if object_session and object_session is not self:
            # Remove the object from the other session, but keep
            # a reference so we can reinstate it.
            object_session.expunge(object_)
            object_._prev_session = weakref.ref(object_session)
        return super().add(object_)

class CarefulAsyncSession(AsyncSession):

    def add(self, object_):
        # Get the session that contains the object, if there is one.
        object_session = async_object_session(object_)
        if object_session and object_session is not self:
            # Remove the object from the other session, but keep
            # a reference so we can reinstate it.
            object_session.expunge(object_)
            object_._prev_session = weakref.ref(object_session)
        return super().add(object_)


# @sa.event.listens_for(CarefulAsyncSession.sync_session, 'after_commit')
def receive_after_commit(session):
    """Reinstate objects into their previous sessions."""
    objects = filter(lambda o: hasattr(o, '_prev_session'), session.identity_map.values())
    for object_ in objects:
        prev_session = object_._prev_session()
        if prev_session:
            session.expunge(object_)
            prev_session.add(object_)
            delattr(object_, '_prev_session')

def receive_after_commit_orm(session):
    """
    Reinstate objects into their previous sessions.

    - This assumes that the object has a `_prev_session` attribute
    """
    objects = filter(lambda o: o._prev_session, session.identity_map.values())
    for object_ in objects:
        prev_session = object_._prev_session()
        if prev_session:
            session.expunge(object_)
            prev_session.add(object_)
            object_._prev_session = None


from ._base import *
from .models import LazyDBCacheBase

LazyDBCacheType = TypeVar("LazyDBCacheType", bound=LazyDBCacheBase)
CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)


from __future__ import annotations

from abc import ABC
from typing import TypeVar, Dict, Callable, Any, Union, Tuple, List, Optional

RT = TypeVar('RT')

# Misc Helpers
LazyDictValue = Union[RT, Callable[[], RT], Tuple[str, Dict[str, Any]], Any]

class LazyDict(ABC):
    """
    This Dict object is different in that 
    it takes str objects and converts them
    to the appropriate type through lazy importing
    """
    _dict: Dict[str, Union[LazyDictValue, Dict[str, LazyDictValue]]] = {}
    excluded_attrs: List[str] = []
    initialize_objects: bool = False

    def __init__(self, *args, **kwargs):
        # Allow initialization by passing keyword arguments or a dictionary
        if len(args) == 1 and not kwargs and isinstance(args[0], dict):
            self._dict = args[0]
        else:
            self._dict = dict(*args, **kwargs)

    def get_or_init(self, name: str, default: Optional[RT] = None) -> RT:
        """
        Get an attribute from the dictionary
        If it does not exist, set it to the default value
        """
        if name not in self._dict:
            if default:
                self._dict[name] = default
            else:
                raise ValueError(f"Default value for {name} is None")
        from lazyops.utils.lazy import lazy_import
        if isinstance(self._dict[name], str):
            self._dict[name] = lazy_import(self._dict[name])
            if self.initialize_objects:
                self._dict[name] = self._dict[name]()
        elif isinstance(self._dict[name], tuple):
            obj_class, kwargs = self._dict[name]
            if isinstance(obj_class, str):
                obj_class = lazy_import(obj_class)
            for k,v in kwargs.items():
                if callable(v): kwargs[k] = v()
            self._dict[name] = obj_class(**kwargs)
        elif isinstance(self._dict[name], dict):
            self._dict[name] = type(self)(self._dict[name])
        return self._dict[name]

    def __getattr__(self, name: str) -> RT:
        """
        Get an attribute from the dictionary        
        """
        if name.startswith('__') and name.endswith('__') or name in self.excluded_attrs:
            return getattr(self, name)
        return self.get_or_init(name, None)

    def __getitem__(self, name: str) -> RT:
        """
        Get an attribute from the dictionary        
        """
        return self.get_or_init(name, None)
    
    def __setitem__(self, name: str, value: RT) -> None:
        """
        Set an attribute in the dictionary        
        """
        self._dict[name] = value
    
    def __contains__(self, name: str) -> bool:
        """
        Check if an attribute is in the dictionary        
        """
        return name in self._dict
    
    def __repr__(self) -> str:
        """
        Get a string representation of the dictionary
        """
        return repr(self._dict)
    
    def __str__(self) -> str:
        """
        Get a string representation of the dictionary
        """
        return str(self._dict)
    
    def __len__(self) -> int:
        """
        Get the length of the dictionary
        """
        return len(self._dict)
    
    def __iter__(self):
        """
        Iterate over the dictionary
        """
        return iter(self._dict)
    
    def keys(self):
        """
        Get the keys of the dictionary
        """
        return self._dict.keys()
    
    def values(self):
        """
        Get the values of the dictionary
        """
        return self._dict.values()
    
    def items(self):
        """
        Get the items of the dictionary
        """
        return self._dict.items()
    
    def get(self, name: str, default: Optional[RT] = None) -> RT:
        """
        Get an attribute from the dictionary
        If it does not exist, set it to the default value
        """
        return self.get_or_init(name, default)
    

LazyDict.register(dict)


from __future__ import annotations

"""
MsgPack Serializer
"""

from typing import Any, Dict, Optional, Union, Type
from lazyops.utils.lazy import lazy_import
from .base import BinaryBaseSerializer, BaseModel, SchemaType, ObjectValue, logger
from ._json import default_json
try:
    import msgpack
    _msgpack_available = True
except ImportError:
    _msgpack_available = False


class MsgPackSerializer(BinaryBaseSerializer):
    name: Optional[str] = "msgpack"
    encoding: Optional[str] = "utf-8"
    disable_object_serialization: Optional[bool] = False
    jsonlib: Any = default_json

    def __init__(
        self, 
        compression: Optional[str] = None,
        compression_level: int | None = None, 
        encoding: str | None = None, 
        serialization_obj: Optional[Type[BaseModel]] = None,
        serialization_obj_kwargs: Optional[Dict[str, Any]] = None,
        disable_object_serialization: Optional[bool] = None,
        **kwargs
    ):
        if not _msgpack_available:
            raise ImportError("MsgPack Serializer is not available. Please install msgpack")
        super().__init__(compression, compression_level, encoding, **kwargs)
        self.serialization_obj = serialization_obj
        self.serialization_obj_kwargs = serialization_obj_kwargs or {}
        self.serialization_schemas: Dict[str, Type[BaseModel]] = {}
        if disable_object_serialization is not None:
            self.disable_object_serialization = disable_object_serialization
        
    def default_serialization_hook(self, obj: ObjectValue):
        """
        Default Serialization Hook
        """
        if not isinstance(obj, BaseModel) and not hasattr(obj, 'model_dump'):
            logger.info(f'Invalid Object Type: |r|{type(obj)}|e| {obj}', colored = True, prefix = "msgpack")
            return obj
        
        if self.disable_object_serialization: 
            return obj.model_dump_json(**self.serialization_obj_kwargs)

        obj_class_name = self.fetch_object_classname(obj)
        if obj_class_name not in self.serialization_schemas:
            self.serialization_schemas[obj_class_name] = obj.__class__
        data = obj.model_dump(mode = 'json', **self.serialization_obj_kwargs)
        data['__class__'] = obj_class_name
        return msgpack.ExtType(2, self.jsonlib.dumps(data).encode(self.encoding))
    
    def default_deserialization_hook(self, code: int, data: Union[str, bytes]) -> ObjectValue:
        """
        Default Deserialization Hook
        """
        if code != 2: return data
        if isinstance(data, bytes): data = data.decode(self.encoding)
        try:
            data = self.jsonlib.loads(data)
        except Exception as e:
            logger.info(f'Error Decoding Value: |r|({type(data)}) {e}|e| {str(data)[:500]}', colored = True, prefix = "msgpack")
            if self.raise_errors: raise e
            return data
        if not self.disable_object_serialization:
            _class = data.pop('__class__', None)
            if _class is not None:
                if _class not in self.serialization_schemas:
                    self.serialization_schemas[_class] = lazy_import(_class)
                _class = self.serialization_schemas[_class]
                return _class.model_validate(data, **self.serialization_obj_kwargs)
        elif self.serialization_obj is not None:
            return self.serialization_obj.model_validate(data, **self.serialization_obj_kwargs)
        return data

    def encode_value(self, value: Union[Any, SchemaType], **kwargs) -> bytes:
        """
        Encode the value with the Pickle Library
        """
        if 'use_bin_type' not in kwargs: kwargs['use_bin_type'] = True
        if 'default' not in kwargs: kwargs['default'] = self.default_serialization_hook
        try:
            return msgpack.packb(value, **kwargs)
        except Exception as e:
            logger.info(f'Error Encoding Value: |r|({type(value)}) {e}|e| {str(value)[:500]}', colored = True, prefix = "msgpack")
            if self.raise_errors: raise e
        return None

    def decode_value(self, value: bytes, **kwargs) -> Union[SchemaType, Dict, Any]:
        """
        Decode the value with the Pickle Library
        """
        try:
            if 'raw' not in kwargs: kwargs['raw'] = False
            if 'ext_hook' not in kwargs: kwargs['ext_hook'] = self.default_deserialization_hook
            return msgpack.unpackb(value, **kwargs)
        except Exception as e:
            logger.info(f'Error Decoding Value: |r|({type(value)}) {e}|e| {str(value)[:500]}', colored = True, prefix = "msgpack")
            if self.raise_errors: raise e
        return None


from __future__ import annotations

from typing import Sequence, Type

from lazyops.libs.dbinit.base import Inspector, TextClause, inspect, text, DeclarativeBase
from lazyops.libs.dbinit.data_structures.grant_to import GrantTo
from lazyops.libs.dbinit.data_structures.privileges import Privilege
from lazyops.libs.dbinit.entities.database import Database
from lazyops.libs.dbinit.entities.database_entity import DatabaseEntity
from lazyops.libs.dbinit.entities.entity import Entity
from lazyops.libs.dbinit.entities.role import Role
from lazyops.libs.dbinit.entities.schema import Schema
from lazyops.libs.dbinit.mixins.grantable import Grantable
from lazyops.libs.dbinit.mixins.sql import SQLBase


class DatabaseContent(DatabaseEntity):
    """
    Represents a `sqlalchemy.orm.DeclarativeBase` and other helpful functionality.
    """

    def __init__(
        self,
        name: str,
        sqlalchemy_base: Type[DeclarativeBase],
        database: Database,
        schemas: Sequence[Schema] | None = None,
        depends_on: Sequence[Entity] | None = None,
        check_if_exists: bool | None = None,
    ):
        """
        :param name: Unique name of the entity. Must be unique within a database.
        :param sqlalchemy_base: The `sqlalchemy.orm.DeclarativeBase` to refer to a collection of tables.
        :param schemas: Any schemas that need to be created prior to the tables specified in the sqlalchemy_base.
        :param database: The :class:`lazyops.libs.dbinit.entities.Database` that this entity belongs to.
        :param depends_on: Any entities that should be created before this one.
        :param check_if_exists: Flag to set existence check behavior. If `True`, will raise an exception during _safe_create if the entity already exists, and will raise an exception during _safe_drop if the entity does not exist.
        """
        super().__init__(name=name, depends_on=depends_on, database=database, check_if_exists=check_if_exists)
        self.base = sqlalchemy_base
        # schemas doesn't do anything since __table_args__ in sqlalchemy defines the schema
        # BUT it helps to have it as a dependency here to remind the user to make schemas they intend to use
        self.schemas = schemas
        self.tables: dict[str, Table] = {
            table.name: Table(name=table.name, database_content=self, schema=table.schema)
            for table in self.base.metadata.tables.values()
        }

    def _create(self) -> None:
        self.base.metadata.create_all(self.database.db_engine())

    def _exists(self) -> bool:
        inspector: Inspector = inspect(self.database.db_engine())
        return all(
            inspector.has_table(table_name=table.name, schema=table.schema)
            for table in self.base.metadata.tables.values()
        )

    def _drop(self) -> None:
        self.base.metadata.drop_all(self.database.db_engine())


class Table(SQLBase, Grantable):
    """
    An internal wrapper for a table, primarily intended to allow easy access to grants.
    """

    def __init__(self, name: str, database_content: DatabaseContent, schema: str | None = "public"):
        """
        :param name: Unique name of the entity. Must be unique within a schema within a database.
        :param database_content: The `lazyops.libs.dbinit.entities.DatabaseContent` this Table belongs to.
        :param schema: The string name of the schema this belongs to.
        """
        super().__init__(name=name)

        if schema:
            self._grant_name = f"{schema}.{name}"

        self.database_content = database_content
        self.schema = schema

    def __hash__(self) -> int:
        return hash((self.name, self.__class__.__name__, self.database_content.name, self.schema))

    def __eq__(self, other: object) -> bool:
        return (
            (
                self.name,
                self.__class__.__name__,
                self.database_content.name,
                self.schema,
            )
            == (
                other.name,
                other.__class__.__name__,
                other.database_content.name,
                self.schema,
            )
            if isinstance(other, self.__class__)
            else NotImplemented
        )

    def grant(self, grants: Sequence[GrantTo]) -> None:
        super().grant(grants=grants)
        self._fix_entity_order(grants=grants, target_entity=self.database_content)

    def _exists(self) -> bool:
        # TODO not sure how expensive creating an inspector is, might not want to do it for every run of this fn
        inspector: Inspector = inspect(self.database_content.database.db_engine())
        return inspector.has_table(table_name=self.name, schema=self.schema)

    def _grant(self, grantee: Role, privileges: set[Privilege]) -> None:
        self._commit_sql(
            engine=self.database_content.database.db_engine(),
            statements=self._grant_statements(grantee=grantee, privileges=privileges),
        )

    def _grants_exist(self, grantee: Role, privileges: set[Privilege]) -> bool:
        rows = self._fetch_sql(
            engine=self.database_content.database.db_engine(), statement=self._grants_exist_statement(grantee=grantee)
        )
        existing_privileges = {r[0] for r in rows}
        return self._check_privileges(declared_privileges=privileges, existing_privileges=existing_privileges)

    def _grants_exist_statement(self, grantee: Role) -> TextClause:
        """
        The SQL statement that checks to see what grants exist.
        :return: A single :class:`sqlalchemy.TextClause` containing the SQL to check what grants exist on this entity.
        """
        return text(
            "SELECT privilege_type FROM information_schema.table_privileges WHERE table_name=:table_name  AND grantee=:grantee_name"
        ).bindparams(table_name=self.name, grantee_name=grantee.name)
        # .bindparams(table_name=self.name, grantee_name=grantee.encoded_name)
        # .bindparams(table_name=self.name, grantee_name=grantee.name)

    def _revoke(self, grantee: Role, privileges: set[Privilege]) -> None:
        self._commit_sql(
            engine=self.database_content.database.db_engine(),
            statements=self._revoke_statements(grantee=grantee, privileges=privileges),
        )

    @staticmethod
    def _allowed_privileges() -> set[Privilege]:
        return {
            Privilege.INSERT,
            Privilege.SELECT,
            Privilege.UPDATE,
            Privilege.DELETE,
            Privilege.TRUNCATE,
            Privilege.REFERENCES,
            Privilege.TRIGGER,
            Privilege.ALL_PRIVILEGES,
        }




"""
Import Handler for slack_sdk
"""

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

try:
    import slack_sdk
    _slack_sdk_available = True
except ImportError:
    slack_sdk = object
    _slack_sdk_available = False

def resolve_slack_sdk(
    required: bool = False,
):
    """
    Ensures that `slack_sdk` is available
    """
    global slack_sdk, _slack_sdk_available
    if not _slack_sdk_available:
        resolve_missing('slack_sdk', required = required)
        import slack_sdk
        _slack_sdk_available = True
        globals()['slack_sdk'] = slack_sdk


def require_slack_sdk(
    required: bool = False,
):
    """
    Wrapper for `resolve_slack_sdk` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_slack_sdk, 
            func = func, 
            required = required
        )
    return decorator


from __future__ import annotations


from .base import BaseCompression, logger
from typing import Optional

try:
    import lz4.frame
    _lz4_available = True
except ImportError:
    _lz4_available = False


class Lz4Compression(BaseCompression):
    name: str = "lz4"
    compression_level: Optional[int] = 4

    def check_deps(self):
        """
        Checks for dependencies
        """
        if _lz4_available is False:
            logger.error("lz4 is not available. Please install lz4 to use lz4 compression")
            raise ImportError("lz4 is not available. Please install lz4 to use lz4 compression")


    def validate_compression_level(self):
        """
        Validates the compression level
        """
        assert self.compression_level in range(17), "Compression level must be between 0 and 16"


    def compress(self, data: bytes, level: Optional[int] = None, **kwargs) -> bytes:
        """
        Compresses the data
        """
        if level is None: level = self.compression_level
        _kwargs = self._compression_kwargs.copy()
        if kwargs: _kwargs.update(kwargs)
        return lz4.frame.compress(data, compression_level = level, **_kwargs)

    def decompress(self, data: bytes, **kwargs) -> bytes:
        """
        Decompresses the data
        """
        _kwargs = self._decompression_kwargs.copy()
        if kwargs: _kwargs.update(kwargs)
        return lz4.frame.decompress(data, **_kwargs)
    
    




from ._base import *

# From https://github.com/peakwinter/python-nginx/blob/master/nginx.py

logger = get_logger('LazyConfig', 'Nginx')

INDENT = '    '

class Error(Exception):
    pass

class ParseError(Error):
    pass

def bump_child_depth(obj, depth):
    children = getattr(obj, 'children', [])
    for child in children:
        child._depth = depth + 1
        bump_child_depth(child, child._depth)


class NginxConfig(object):
    """
    Represents an nginx configuration.
    A `NginxConfig` can consist of any number of server blocks, as well as Upstream
    and other types of containers. It can also include top-level comments.
    """

    def __init__(self, *args):
        """
        Initialize object.
        :param *args: Any objects to include in this NginxConfig.
        """
        self.children = list(args)

    def add(self, *args):
        """
        Add object(s) to the NginxConfig.
        :param *args: Any objects to add to the NginxConfig.
        :returns: full list of NginxConfig's child objects
        """
        self.children.extend(args)
        return self.children

    def remove(self, *args):
        """
        Remove object(s) from the NginxConfig.
        :param *args: Any objects to remove from the NginxConfig.
        :returns: full list of NginxConfig's child objects
        """
        for x in args:
            self.children.remove(x)
        return self.children

    def filter(self, btype='', name=''):
        """
        Return child object(s) of this NginxConfig that satisfy certain criteria.
        :param str btype: Type of object to filter by (e.g. 'Key')
        :param str name: Name of key OR container value to filter by
        :returns: full list of matching child objects
        """
        filtered = []
        for x in self.children:
            if name and isinstance(x, Key) and x.name == name:
                filtered.append(x)
            elif isinstance(x, Container) and x.__class__.__name__ == btype \
                    and x.value == name:
                filtered.append(x)
            elif not name and btype and x.__class__.__name__ == btype:
                filtered.append(x)
        return filtered

    @property
    def servers(self):
        """Return a list of child Server objects."""
        return [x for x in self.children if isinstance(x, Server)]

    @property
    def server(self):
        """Convenience property to fetch the first available server only."""
        return self.servers[0]

    @property
    def as_list(self):
        """Return all child objects in nested lists of strings."""
        return [x.as_list for x in self.children]

    @property
    def as_dict(self):
        """Return all child objects in nested dict."""
        return {'conf': [x.as_dict for x in self.children]}

    @property
    def as_strings(self):
        """Return the entire NginxConfig as nginx config strings."""
        ret = []
        for x in self.children:
            if isinstance(x, (Key, Comment)):
                ret.append(x.as_strings)
            else:
                for y in x.as_strings:
                    ret.append(y)
        if ret:
            ret[-1] = re.sub('}\n+$', '}\n', ret[-1])
        return ret


class Container(object):
    """
    Represents a type of child block found in an nginx config.
    Intended to be subclassed by various types of child blocks, like
    Locations or Geo blocks.
    """

    def __init__(self, value, *args):
        """
        Initialize object.
        :param str value: Value to be used in name (e.g. regex for Location)
        :param *args: Any objects to include in this NginxConfig.
        """
        self.name = ''
        self.value = value
        self._depth = 0
        self.children = list(args)
        bump_child_depth(self, self._depth)

    def add(self, *args):
        """
        Add object(s) to the Container.
        :param *args: Any objects to add to the Container.
        :returns: full list of Container's child objects
        """
        self.children.extend(args)
        bump_child_depth(self, self._depth)
        return self.children

    def remove(self, *args):
        """
        Remove object(s) from the Container.
        :param *args: Any objects to remove from the Container.
        :returns: full list of Container's child objects
        """
        for x in args:
            self.children.remove(x)
        return self.children

    def filter(self, btype='', name=''):
        """
        Return child object(s) of this Server block that meet certain criteria.
        :param str btype: Type of object to filter by (e.g. 'Key')
        :param str name: Name of key OR container value to filter by
        :returns: full list of matching child objects
        """
        filtered = []
        for x in self.children:
            if name and isinstance(x, Key) and x.name == name:
                filtered.append(x)
            elif isinstance(x, Container) and x.__class__.__name__ == btype \
                    and x.value == name:
                filtered.append(x)
            elif not name and btype and x.__class__.__name__ == btype:
                filtered.append(x)
        return filtered

    @property
    def locations(self):
        """Return a list of child Location objects."""
        return [x for x in self.children if isinstance(x, Location)]

    @property
    def comments(self):
        """Return a list of child Comment objects."""
        return [x for x in self.children if isinstance(x, Comment)]

    @property
    def keys(self):
        """Return a list of child Key objects."""
        return [x for x in self.children if isinstance(x, Key)]

    @property
    def as_list(self):
        """Return all child objects in nested lists of strings."""
        return [self.name, self.value, [x.as_list for x in self.children]]

    @property
    def as_dict(self):
        """Return all child objects in nested dict."""
        dicts = [x.as_dict for x in self.children]
        return {'{0} {1}'.format(self.name, self.value): dicts}

    @property
    def as_strings(self):
        """Return the entire Container as nginx config strings."""
        ret = []
        container_title = (INDENT * self._depth)
        container_title += '{0}{1} {{\n'.format(
            self.name, (' {0}'.format(self.value) if self.value else '')
        )
        ret.append(container_title)
        for x in self.children:
            if isinstance(x, Key):
                ret.append(INDENT + x.as_strings)
            elif isinstance(x, Comment):
                if x.inline and len(ret) >= 1:
                    ret[-1] = ret[-1].rstrip('\n') + '  ' + x.as_strings
                else:
                    ret.append(INDENT + x.as_strings)
            elif isinstance(x, Container):
                y = x.as_strings
                ret.append('\n' + y[0])
                for z in y[1:]:
                    ret.append(INDENT + z)
            else:
                y = x.as_strings
                ret.append(INDENT + y)
        ret[-1] = re.sub('}\n+$', '}\n', ret[-1])
        ret.append('}\n\n')
        return ret


class Comment(object):
    """Represents a comment in an nginx config."""

    def __init__(self, comment, inline=False):
        """
        Initialize object.
        :param str comment: Value of the comment
        :param bool inline: This comment is on the same line as preceding item
        """
        self.comment = comment
        self.inline = inline

    @property
    def as_list(self):
        """Return comment as nested list of strings."""
        return [self.comment]

    @property
    def as_dict(self):
        """Return comment as dict."""
        return {'#': self.comment}

    @property
    def as_strings(self):
        """Return comment as nginx config string."""
        return '# {0}\n'.format(self.comment)


class Http(Container):
    """Container for HTTP sections in the main NGINX conf file."""

    def __init__(self, *args):
        """Initialize."""
        super(Http, self).__init__('', *args)
        self.name = 'http'


class Server(Container):
    """Container for server block configurations."""

    def __init__(self, *args):
        """Initialize."""
        super(Server, self).__init__('', *args)
        self.name = 'server'

    @property
    def as_dict(self):
        """Return all child objects in nested dict."""
        return {'server': [x.as_dict for x in self.children]}


class Location(Container):
    """Container for Location-based options."""

    def __init__(self, value, *args):
        """Initialize."""
        super(Location, self).__init__(value, *args)
        self.name = 'location'


class Events(Container):
    """Container for Event-based options."""

    def __init__(self, *args):
        """Initialize."""
        super(Events, self).__init__('', *args)
        self.name = 'events'


class LimitExcept(Container):
    """Container for specifying HTTP method restrictions."""

    def __init__(self, value, *args):
        """Initialize."""
        super(LimitExcept, self).__init__(value, *args)
        self.name = 'limit_except'


class Types(Container):
    """Container for MIME type mapping."""

    def __init__(self, *args):
        """Initialize."""
        super(Types, self).__init__('', *args)
        self.name = 'types'


class If(Container):
    """Container for If conditionals."""

    def __init__(self, value, *args):
        """Initialize."""
        super(If, self).__init__(value, *args)
        self.name = 'if'


class Upstream(Container):
    """Container for upstream configuration (reverse proxy)."""

    def __init__(self, value, *args):
        """Initialize."""
        super(Upstream, self).__init__(value, *args)
        self.name = 'upstream'


class Geo(Container):
    """
    Container for geo module configuration.
    See docs here: http://nginx.org/en/docs/http/ngx_http_geo_module.html
    """

    def __init__(self, value, *args):
        """Initialize."""
        super(Geo, self).__init__(value, *args)
        self.name = 'geo'


class Map(Container):
    """Container for map configuration."""

    def __init__(self, value, *args):
        """Initialize."""
        super(Map, self).__init__(value, *args)
        self.name = 'map'


class Stream(Container):
    """Container for stream sections in the main NGINX conf file."""

    def __init__(self, *args):
        """Initialize."""
        super(Stream, self).__init__('', *args)
        self.name = 'stream'


class Key(object):
    """Represents a simple key/value object found in an nginx config."""

    def __init__(self, name, value):
        """
        Initialize object.
        :param *args: Any objects to include in this Server block.
        """
        self.name = name
        self.value = value

    @property
    def as_list(self):
        """Return key as nested list of strings."""
        return [self.name, self.value]

    @property
    def as_dict(self):
        """Return key as dict key/value."""
        return {self.name: self.value}

    @property
    def as_strings(self):
        """Return key as nginx config string."""
        if self.value == '' or self.value is None:
            return '{0};\n'.format(self.name)
        if type(self.value) == str and '"' not in self.value and (';' in self.value or '#' in self.value):
            return '{0} "{1}";\n'.format(self.name, self.value)
        return '{0} {1};\n'.format(self.name, self.value)


class Nginx:

    @classmethod
    def loads(cls, data, conf=True):
        """
        Load an nginx configuration from a provided string.
        :param str data: nginx configuration
        :param bool conf: Load object(s) into a NginxConfig object?
        """
        f = NginxConfig() if conf else []
        lopen = []
        index = 0

        while True:
            m = re.compile(r'^\s*events\s*{').search(data[index:])
            if m:
                logger.debug("Open (Events)")
                e = Events()
                lopen.insert(0, e)
                index += m.end()
                continue

            m = re.compile(r'^\s*http\s*{').search(data[index:])
            if m:
                logger.debug("Open (Http)")
                h = Http()
                lopen.insert(0, h)
                index += m.end()
                continue

            m = re.compile(r'^\s*stream\s*{').search(data[index:])
            if m:
                logger.debug("Open (Stream)")
                s = Stream()
                lopen.insert(0, s)
                index += m.end()
                continue

            m = re.compile(r'^\s*server\s*{').search(data[index:])
            if m:
                logger.debug("Open (Server)")
                s = Server()
                lopen.insert(0, s)
                index += m.end()
                continue

            n = re.compile(r'(?!\B"[^"]*);(?![^"]*"\B)')
            m = re.compile(r'^\s*location\s+(.*?".*?".*?|.*?)\s*{').search(data[index:])
            if m and not n.search(m.group()):
                logger.debug("Open (Location) {0}".format(m.group(1)))
                l = Location(m.group(1))
                lopen.insert(0, l)
                index += m.end()
                continue

            m = re.compile(r'^\s*if\s+(.*?".*?".*?|.*?)\s*{').search(data[index:])
            if m and not n.search(m.group()):
                logger.debug("Open (If) {0}".format(m.group(1)))
                ifs = If(m.group(1))
                lopen.insert(0, ifs)
                index += m.end()
                continue

            m = re.compile(r'^\s*upstream\s+(.*?)\s*{').search(data[index:])
            if m and not n.search(m.group()):
                logger.debug("Open (Upstream) {0}".format(m.group(1)))
                u = Upstream(m.group(1))
                lopen.insert(0, u)
                index += m.end()
                continue

            m = re.compile(r'^\s*geo\s+(.*?".*?".*?|.*?)\s*{').search(data[index:])
            if m and not n.search(m.group()):
                logger.debug("Open (Geo) {0}".format(m.group(1)))
                g = Geo(m.group(1))
                lopen.insert(0, g)
                index += m.end()
                continue

            m = re.compile(r'^\s*map\s+(.*?".*?".*?|.*?)\s*{').search(data[index:])
            if m and not n.search(m.group()):
                logger.debug("Open (Map) {0}".format(m.group(1)))
                g = Map(m.group(1))
                lopen.insert(0, g)
                index += m.end()
                continue

            m = re.compile(r'^\s*limit_except\s+(.*?".*?".*?|.*?)\s*{').search(data[index:])
            if m and not n.search(m.group()):
                logger.debug("Open (LimitExcept) {0}".format(m.group(1)))
                l = LimitExcept(m.group(1))
                lopen.insert(0, l)
                index += m.end()
                continue

            m = re.compile(r'^\s*types\s*{').search(data[index:])
            if m:
                logger.debug("Open (Types)")
                l = Types()
                lopen.insert(0, l)
                index += m.end()
                continue

            m = re.compile(r'^(\s*)#[ \r\t\f]*(.*?)\n').search(data[index:])
            if m:
                logger.debug("Comment ({0})".format(m.group(2)))
                c = Comment(m.group(2), inline='\n' not in m.group(1))
                if lopen and isinstance(lopen[0], Container):
                    lopen[0].add(c)
                else:
                    f.add(c) if conf else f.append(c)
                index += m.end() - 1
                continue

            m = re.compile(r'^\s*}').search(data[index:])
            if m:
                if isinstance(lopen[0], Container):
                    logger.debug("Close ({0})".format(lopen[0].__class__.__name__))
                    c = lopen[0]
                    lopen.pop(0)
                    if lopen and isinstance(lopen[0], Container):
                        lopen[0].add(c)
                    else:
                        f.add(c) if conf else f.append(c)
                index += m.end()
                continue

            if ";" not in data[index:] and "}" in data[index:]:
                # If there is still something to parse, expect ';' otherwise
                # the Key regexp can get stuck due to regexp catastrophic backtracking
                raise ParseError("Config syntax, missing ';' at index: {}".format(index))

            double = r'\s*"[^"]*"'
            single = r'\s*\'[^\']*\''
            normal = r'\s*[^;\s]*'
            s1 = r'{}|{}|{}'.format(double, single, normal)
            s = r'^\s*({})\s*((?:{})+);'.format(s1, s1)
            m = re.compile(s).search(data[index:])
            if m:
                logger.debug("Key {0} {1}".format(m.group(1), m.group(2)))
                k = Key(m.group(1), m.group(2))
                if lopen and isinstance(lopen[0], (Container, Server)):
                    lopen[0].add(k)
                else:
                    f.add(k) if conf else f.append(k)
                index += m.end()
                continue

            m = re.compile(r'^\s*(\S+);').search(data[index:])
            if m:
                logger.debug("Key {0}".format(m.group(1)))
                k = Key(m.group(1), '')
                if lopen and isinstance(lopen[0], (Container, Server)):
                    lopen[0].add(k)
                else:
                    f.add(k) if conf else f.append(k)
                index += m.end()
                continue

            break

        return f


    @classmethod
    def load(cls, pathlike_or_filestr):
        """
        Load an nginx configuration from a provided file-like object.
        :param obj fobj: nginx configuration
        """
        data = PathIO(pathlike_or_filestr).read_text() if isinstance(pathlike_or_filestr, str) else pathlike_or_filestr.read()
        return Nginx.loads(data)

    @classmethod
    def dumps(cls, obj):
        """
        Dump an nginx configuration to a string.
        :param obj obj: nginx object (NginxConfig, Server, Container)
        :returns: nginx configuration as string
        """
        return ''.join(obj.as_strings)

    @classmethod
    def dump(cls, obj, pathlike_or_filestr):
        """
        Write an nginx configuration to a file-like object.
        :param obj obj: nginx object (NginxConfig, Server, Container)
        :param obj fobj: file-like object to write to
        :returns: file-like object that was written to
        """
        pathlike = PathIO(pathlike_or_filestr) if isinstance(pathlike_or_filestr, str) else pathlike_or_filestr
        pathlike.write(Nginx.dumps(obj))
        return pathlike



from __future__ import annotations

import json
from typing import Any, Dict, Optional, Union, Type
from lazyops.utils.lazy import lazy_import
from .base import BaseSerializer, ObjectValue, SchemaType, BaseModel, logger

try:
    import orjson
    _orjson_available = True
except ImportError:
    _orjson_available = False

try:
    import simdjson
    _simdjson_available = True
except ImportError:
    _simdjson_available = False

try:
    import ujson
    _ujson_available = True
except ImportError:
    _ujson_available = False


if _simdjson_available:
    default_json = simdjson

elif _orjson_available:
    default_json = orjson

elif _ujson_available:
    default_json = ujson

else:
    default_json = json


class JsonSerializer(BaseSerializer):

    name: Optional[str] = "json"
    encoding: Optional[str] = "utf-8"
    jsonlib: Any = default_json
    disable_object_serialization: Optional[bool] = False

    def __init__(
        self, 
        jsonlib: Optional[Union[str, Any]] = None,
        compression: Optional[str] = None,
        compression_level: int | None = None, 
        encoding: str | None = None, 
        serialization_obj: Optional[Type[BaseModel]] = None,
        serialization_obj_kwargs: Optional[Dict[str, Any]] = None,
        disable_object_serialization: Optional[bool] = None,
        **kwargs
    ):
        super().__init__(compression, compression_level, encoding, **kwargs)
        self.serialization_obj = serialization_obj
        self.serialization_obj_kwargs = serialization_obj_kwargs or {}
        self.serialization_schemas: Dict[str, Type[BaseModel]] = {}
        if disable_object_serialization is not None:
            self.disable_object_serialization = disable_object_serialization
        if jsonlib is not None:
            if isinstance(jsonlib, str):
                jsonlib = lazy_import(jsonlib)
            assert hasattr(jsonlib, "dumps") and hasattr(jsonlib, "loads"), f"Invalid JSON Library: {jsonlib}"
            self.jsonlib = jsonlib
        self.jsonlib_name = self.jsonlib.__name__
    

    def encode_one(self, value: Union[Any, SchemaType], **kwargs) -> str:
        """
        Encode the value with the JSON Library
        """
        try:
            if hasattr(value, 'model_dump'): 
                if not self.disable_object_serialization:
                    obj_class_name = self.fetch_object_classname(value)
                    if obj_class_name not in self.serialization_schemas:
                        self.serialization_schemas[obj_class_name] = value.__class__
                value = value.model_dump(mode = 'json', round_trip = True, **self.serialization_obj_kwargs)
                if not self.disable_object_serialization:
                    value['__class__'] = obj_class_name
        except Exception as e:
            logger.info(f'Error Encoding Value: |r|({type(value)}) {e}|e| {str(value)[:1000]}', colored = True)
        try:
            return self.jsonlib.dumps(value, **kwargs)
        except Exception as e:
            logger.info(f'Error Encoding Value: |r|({type(value)}) {e}|e| {str(value)[:1000]}', colored = True, prefix = self.jsonlib_name)
            if self.raise_errors: raise e
        return None
        
    def encode_value(self, value: Union[Any, SchemaType], **kwargs) -> str:
        """
        Encode the value with the JSON Library
        """
        if isinstance(value, list):
            values = [self.encode_one(v, **kwargs) for v in value]
            value_dict = {
                '__type__': 'list',
                'values': values,
            }
            return self.jsonlib.dumps(value_dict, **kwargs)
        return self.encode_one(value, **kwargs)


    def decode_one(self, value: str, **kwargs) -> Union[SchemaType, Dict, Any]:
        """
        Decode the value with the JSON Library
        """
        try:
            value = self.jsonlib.loads(value, **kwargs)
            if not self.disable_object_serialization and isinstance(value, dict) and '__class__' in value:
                obj_class_name = value.pop('__class__')
                if obj_class_name not in self.serialization_schemas:
                    self.serialization_schemas[obj_class_name] = lazy_import(obj_class_name)
                obj_class = self.serialization_schemas[obj_class_name]
                value = obj_class.model_validate(value)
            elif self.serialization_obj is not None:
                value = self.serialization_obj.model_validate(value)
            return value
        except Exception as e:
            logger.info(f'Error Decoding Value: |r|({type(value)}) {e}|e| {str(value)[:1000]}', colored = True, prefix = self.jsonlib_name)
            if self.raise_errors: raise e
        return None

    def decode_value(self, value: str, **kwargs) -> Union[SchemaType, Dict, Any]:
        """
        Decode the value with the JSON Library
        """
        value = self.decode_one(value, **kwargs)
        if isinstance(value, dict) and '__type__' in value and value['__type__'] == 'list':
            return [self.decode_one(v, **kwargs) for v in value['values']]
        return value

        
    
    

    







import datetime

def convert_date_to_rfc3339(date: str) -> str:
    """
    Converts a date to RFC3339 format, as Weaviate requires dates to be in RFC3339 format including the time and
    timezone.

    If the provided date string does not contain a time and/or timezone, we use 00:00 as default time
    and UTC as default time zone.
    """
    parsed_datetime = datetime.datetime.fromisoformat(date)
    return (
        f"{parsed_datetime.isoformat()}Z"
        if parsed_datetime.utcoffset() is None
        else parsed_datetime.isoformat()
    )



from __future__ import annotations

"""
Type Helpers
"""


import uuid
import datetime
import contextlib
from enum import Enum
from typing import Dict, Any, List, Optional, Union, Type
from pydantic import Field, BaseModel
from sqlalchemy.types import TypeDecorator, Text, String, VARCHAR
from sqlalchemy.sql import operators
from sqlalchemy.dialects.postgresql import JSONB
from lazyops.utils import create_unique_id, create_timestamp, timer, Json

DatetimeT = datetime.datetime
OptionalDatetimeT = Optional[datetime.datetime]

UUIDT = uuid.UUID
OptionalUUIDT = Optional[uuid.UUID]

UniqueIDT = str
OptionalUniqueIDT = Optional[str]

DurationT = float
OptionalDurationT = Optional[float]


# Factories
DurationFactory = lambda: Field(default_factory = timer)
DateTimeFactory = lambda: Field(default_factory = create_timestamp)
UniqueIDFactory = lambda: Field(default_factory = create_unique_id)

NullAlias = lambda x: Field(None, alias = x)

# Enums
class OrderBy(Enum):
    ASC = 'asc'
    DESC = 'desc'


class PaginationParams(BaseModel):
    """Pagination params for endpoints."""
    offset: int = 0
    limit: Optional[int] = 100
    order_by: Optional[Any] = None # OrderBy.asc


try:
    from fileio import File, FileLike

    class FileField(TypeDecorator):

        impl = Text
        cache_ok = True

        def coerce_compared_value(self, op, value):
            return Text() if op in (operators.like_op, operators.not_like_op) else self

        def process_bind_param(self, value: Optional[Union[str, FileLike]], dialect):
            """
            Return the value to be stored in the database.
            """
            if value is None: return None
            if isinstance(value, str): return value
            return value.as_posix() \
                if hasattr(value, 'as_posix') \
                    else str(value)

        def process_result_value(self, value: Optional[Union[str, Any]], dialect):
            return File(value) if value is not None else None

except ImportError:
    FileField: object = None

# https://amercader.net/blog/beware-of-json-fields-in-sqlalchemy/


class JsonString(TypeDecorator):
    """Enables JSON storage by encoding and decoding on the fly."""

    impl = String
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is not None: value = Json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None: value = Json.loads(value)
        return value
    
    def coerce_compared_value(self, op, value):
        return Text() \
            if op in (operators.like_op, operators.not_like_op) \
                else self


try:
    from sqlalchemy_json import mutable_json_type, NestedMutable, MutableDict

    JsonStringField: Type[MutableDict] = mutable_json_type(dbtype = JsonString, nested=True)
    JsonStringNestedField: Type[NestedMutable] = mutable_json_type(dbtype = JsonString, nested = True)

    JSONBField: Type[MutableDict] = mutable_json_type(dbtype = JSONB)
    JSONBNestedField: Type[NestedMutable] = mutable_json_type(dbtype = JSONB, nested = True)


except ImportError:
    JsonStringField = JsonString
    JsonStringNestedField = JsonString

    JSONBField = JSONB
    JSONBNestedField = JSONB



"""
K8s Configs
"""
import pathlib

from typing import Optional, Union
from lazyops.types.models import BaseSettings, validator
from lazyops.types.classprops import lazyproperty

class K8sSettings(BaseSettings):
    namespace: Optional[str] = None
    kubeconfig: Optional[Union[str, pathlib.Path]] = None
    kubernetes_service_host: Optional[str] = None
    kubernetes_service_port: Optional[int] = None

    @validator("kubeconfig", pre=True)
    def validate_kubeconfig(cls, v):
        if v is not None: return v
        from lazyops.utils.system import (
            get_k8s_kubeconfig
        )
        return get_k8s_kubeconfig()
    
    @validator("namespace", pre=True)
    def validate_namespace(cls, v):
        if v is not None: return v
        from lazyops.utils.system import (
            get_k8s_namespace
        )
        return get_k8s_namespace()

    @lazyproperty
    def in_k8s(self) -> bool:
        from lazyops.utils.system import (
            is_in_kubernetes
        )
        return is_in_kubernetes()
    
    @lazyproperty
    def in_cluster(self) -> bool:
        return (
            self.kubernetes_service_host is not None and self.kubernetes_service_port is not None
        ) or self.in_k8s
    



from __future__ import annotations

from typing import Sequence

from lazyops.libs.dbinit.entities.database import Database
from lazyops.libs.dbinit.entities.entity import Entity
from lazyops.libs.dbinit.mixins.sql import SQLCreatable


class DatabaseEntity(Entity):
    """
    Parent class for database-wide entities (most everything that isn't a database or role).
    """

    def __init__(
        self,
        name: str,
        database: Database,
        depends_on: Sequence[Entity] | None = None,
        check_if_exists: bool | None = None,
    ):
        """
        :param name: Unique name of the entity. Cluster-level entities must be unique, database-level entities must be unique within a database.
        :param database: The :class:`lazyops.libs.dbinit.entities.Database` that this entity belongs to.
        :param depends_on: Any entities that should be created before this one.
        :param check_if_exists: Flag to set existence check behavior. If `True`, will raise an exception during _safe_create if the entity already exists, and will raise an exception during _safe_drop if the entity does not exist.
        """
        self.database = database
        super().__init__(name=name, depends_on=depends_on, check_if_exists=check_if_exists)

    def __hash__(self) -> int:
        return hash((self.name, self.__class__.__name__, self.database.name))

    def __eq__(self, other: object) -> bool:
        return (
            (self.name, self.__class__.__name__, self.database.name)
            == (
                other.name,
                other.__class__.__name__,
                other.database.name,
            )
            if isinstance(other, self.__class__)
            else NotImplemented
        )


class DatabaseSqlEntity(SQLCreatable, DatabaseEntity):
    """
    Parent class for creatable, database-wide entities. Defines some of the abstract methods needed since they are
    consistent across entities.
    """

    def _create(self) -> None:
        self._commit_sql(engine=self.database.db_engine(), statements=self._create_statements())

    def _exists(self) -> bool:
        rows = self._fetch_sql(engine=self.database.db_engine(), statement=self._exists_statement())
        return rows[0][0]  # type: ignore

    def _drop(self) -> None:
        self._commit_sql(engine=self.database.db_engine(), statements=self._drop_statements())



"""
Support Lazy Loading of Modules

- Credit: https://github.com/kevdog824
"""

from .main import LazyLoad, lazy_load, load
from .wrappers import lazy_function_wrapper
from typing import TYPE_CHECKING


import os
import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings('ignore', message='Unverified HTTPS request')
warnings.filterwarnings('ignore', message='RequestsDependencyWarning')


lazyops_root = os.path.abspath(os.path.dirname(__file__))
from .lazyclasses import lazyclass
from dataclasses import dataclass
import typing as types

from .envs import LazyEnv, get_logger, lazywatcher, lazywatch
from .models import LazyData, LazyTime, LazyDate, LazyFormatter, LazyTimer, LazyObject
from .common import lazylibs, lazy_init, run_cmd, clone_repo, File, PathIO, PathIOLike, get_pathlike
from .common import autojson, read_json, read_jsonlines

from .utils import find_binary_in_path, timed_cache, latest_tf_ckpt, require_module
from .utils import build_chunks, retryable, list_to_dict, create_uuid
from .mp import lazy_parallelize, lazyproc, lazymultiproc, LazyProcs, LazyProc, async_to_sync
from .apis import LazyAPI, LazyAPIConfig


lazyenv = LazyEnv
lazyitem = LazyData
timer = LazyTime
timers = LazyTimer
formatter = LazyFormatter
tstamp = LazyDate.dtime
dtime = LazyDate
ddate = LazyDate.date
fio = File
lazyapi = LazyAPI
lazyapiconfig = LazyAPIConfig
lazy_async = async_to_sync

from .lazyio import LazyHFModel
from . import lazyconfig

from .serializers import async_cache

__all__ = [
    'types',
    'dataclass',
    'lazyclass',
    'LazyEnv',
    'get_logger',
    'LazyData',
    'LazyTime',
    'LazyDate',
    'LazyFormatter',
    'LazyTimer',
    'LazyObject',
    'LazyHFModel',
    'lazylibs',
    'lazy_init',
    'run_cmd',
    'clone_repo',
    'File',
    'lazyenv',
    'lazyitem',
    'timer',
    'timers',
    'formatter',
    'dtime',
    'tstamp',
    'ddate',
    'fio',
    'lazyapi',
    'lazyapiconfig',
    'lazywatch',
    'lazywatcher',
    'lazy_parallelize',
    'lazyproc',
    'lazymultiproc',
    'LazyProcs',
    'LazyProc',
    'lazy_async',
    'async_to_sync',
    'async_cache',
    'PathIO',
    'PathIOLike', 
    'get_pathlike'
    'lazyconfig'
]
#from . import lazyrpc



"""
Import Handler for `transformers`
"""

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

try:
    import transformers
    _transformers_available = True
except ImportError:
    transformers = object
    _transformers_available = False

def resolve_transformers(
    required: bool = False,
    version: str = None,
):
    """
    Ensures that `transformers` is available
    """
    global transformers, _transformers_available
    if not _transformers_available:
        pkg = 'transformers'
        if version is not None:
            pkg += f'=={version}'
        resolve_missing(pkg, required = required)
        import transformers
        _transformers_available = True
        globals()['transformers'] = transformers


def require_transformers(
    required: bool = False,
    version: str = None,
):
    """
    Wrapper for `resolve_transformers` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_transformers, 
            func = func, 
            required = required,
            version = version
        )
    return decorator


"""
Async Helpers
"""
import sys
import asyncio
import functools

from typing import Optional, Any, Callable, Awaitable, Union, TypeVar, Coroutine, Iterable, AsyncIterable, AsyncIterator, AsyncGenerator
from lazyops.utils.system import get_cpu_count
from lazyops.utils.pooler import ThreadPooler
from lazyops.types.common import UpperStrEnum

if sys.version_info < (3, 10):
    # Add aiter and anext to asyncio
    def aiter(it: AsyncIterable) -> Any:
        return it.__aiter__()
    
    def anext(it: AsyncIterator) -> Any:
        return it.__anext__()
    

_concurrency_limit: Optional[int] = None
RT = TypeVar("RT")

def set_concurrency_limit(
    limit: Optional[int] = None
):
    """
    Set the concurrency limit
    """
    global _concurrency_limit
    if limit is None: limit = get_cpu_count() * 4
    _concurrency_limit = limit
    # asyncio.set_event_loop_policy(asyncio.BoundedSemaphore(limit))

def get_concurrency_limit() -> Optional[int]:
    """
    Get the concurrency limit
    """
    if _concurrency_limit is None: set_concurrency_limit()
    return _concurrency_limit

# ref
# https://death.andgravity.com/limit-concurrency

class ReturnWhenType(UpperStrEnum):
    """
    Return When Type
    """
    FIRST_COMPLETED = "FIRST_COMPLETED"
    FIRST_EXCEPTION = "FIRST_EXCEPTION"
    ALL_COMPLETED = "ALL_COMPLETED"

    @property
    def val(self) -> Union[asyncio.FIRST_COMPLETED, asyncio.FIRST_EXCEPTION, asyncio.ALL_COMPLETED]:
        """
        Get the value of the return when type
        """
        return getattr(asyncio, self.value)

async def limit_concurrency(
    mapped_iterable: Union[Callable[[], Awaitable[Any]], Awaitable[Any], Coroutine[Any, Any, Any], Callable[[], Any]],
    limit: Optional[int] = None,
    return_when: Optional[ReturnWhenType] = ReturnWhenType.FIRST_COMPLETED,
):
    """
    Limit the concurrency of an iterable

    Args:
        mapped_iterable (Union[Callable[[], Awaitable[Any]], Awaitable[Any], Coroutine[Any, Any, Any], Callable[[], Any]]): The iterable to limit the concurrency of
        limit (Optional[int], optional): The limit of the concurrency. Defaults to None.
        return_when (Optional[ReturnWhenType], optional): The return when type. Defaults to ReturnWhenType.FIRST_COMPLETED.
    
    Yields:
        [type]: [description]
    """
    try:
        iterable = aiter(mapped_iterable)
        is_async = True
    except (TypeError, AttributeError):
        iterable = iter(mapped_iterable)
        is_async = False
    
    iterable_ended: bool = False
    pending = set()
    limit = get_concurrency_limit() if limit is None else limit
    return_when = ReturnWhenType(return_when) if isinstance(return_when, str) else return_when

    while pending or not iterable_ended:
        while len(pending) < limit and not iterable_ended:
            try:
                iter_item = await anext(iterable) if is_async else next(iterable)
            except StopAsyncIteration if is_async else StopIteration:
                iterable_ended = True
            else:
                pending.add(asyncio.ensure_future(iter_item))

        if not pending: return
        done, pending = await asyncio.wait(
            pending, 
            return_when = return_when.val
        )
        while done: yield done.pop()

def ensure_coro(
    func: Callable[..., Any]
) -> Callable[..., Awaitable[Any]]:
    """
    Ensure that the function is a coroutine
    """
    if asyncio.iscoroutinefunction(func): return func
    @functools.wraps(func)
    async def inner(*args, **kwargs):
        return await ThreadPooler.asyncish(func, *args, **kwargs)
    return inner


async def amap(
    func: Callable[..., Awaitable[Any]],
    iterable: Iterable[Any], 
    *, 
    limit: Optional[int] = None,
    return_when: Optional[ReturnWhenType] = ReturnWhenType.FIRST_COMPLETED,
):
    """
    Async Map

    Args:
        func (Callable[..., Awaitable[Any]]): The function to map
        iterable (Iterable[Any]): The iterable to map
        limit (Optional[int], optional): The limit of the concurrency. Defaults to None.
        return_when (Optional[ReturnWhenType], optional): The return when type. Defaults to ReturnWhenType.FIRST_COMPLETED.
    
    Yields:
        [type]: [description]
    """
    func = ensure_coro(func)
    try:
        mapped_iterable = map(func, iterable)
    except TypeError:
        mapped_iterable = (func(x) async for x in iterable)
    async for task in limit_concurrency(mapped_iterable, limit = limit, return_when = return_when):
        yield await task

async def amap_unordered(
    func: Callable[..., Awaitable[Any]],
    iterable: Iterable[Any], 
    *, 
    limit: Optional[int] = None,
    return_when: Optional[ReturnWhenType] = ReturnWhenType.FIRST_COMPLETED,
):
    """
    Async Map Unordered

    Args:
        func (Callable[..., Awaitable[Any]]): The function to map
        iterable (Iterable[Any]): The iterable to map
        limit (Optional[int], optional): The limit of the concurrency. Defaults to None.
        return_when (Optional[ReturnWhenType], optional): The return when type. Defaults to ReturnWhenType.FIRST_COMPLETED.
    
    Yields:
        [type]: [description]
    """
    async for result in amap(func, iterable, limit = limit, return_when = return_when):
        yield result


async def amap_ordered(
    func: Callable[..., Awaitable[Any]],
    iterable: Iterable[Any], 
    *, 
    limit: Optional[int] = None,
    return_when: Optional[ReturnWhenType] = ReturnWhenType.ALL_COMPLETED,
):
    """
    Async Map Ordered

    Args:
        func (Callable[..., Awaitable[Any]]): The function to map
        iterable (Iterable[Any]): The iterable to map
        limit (Optional[int], optional): The limit of the concurrency. Defaults to None.
        return_when (Optional[ReturnWhenType], optional): The return when type. Defaults to ReturnWhenType.FIRST_COMPLETED.
    
    Yields:
        [type]: [description]
    """
    async for result in amap(func, iterable, limit = limit, return_when = return_when):
        yield result


async def amap_v2(
    func: Callable[..., Awaitable[Any]],
    iterable: Iterable[Any], 
    *args,
    limit: Optional[int] = None,
    return_when: Optional[ReturnWhenType] = ReturnWhenType.FIRST_COMPLETED,
    **kwargs,
):
    """
    Async Map

    Args:
        func (Callable[..., Awaitable[Any]]): The function to map
        iterable (Iterable[Any]): The iterable to map
        limit (Optional[int], optional): The limit of the concurrency. Defaults to None.
        return_when (Optional[ReturnWhenType], optional): The return when type. Defaults to ReturnWhenType.FIRST_COMPLETED.
    
    Yields:
        [type]: [description]
    """
    func = ensure_coro(func)
    # Deal with Tuples and Lists in the iterable
    # if isinstance(iterable, (tuple, list)):
    #     iterable = [func(x, *args, **kwargs) for x in iterable]
    # else:
    #     iterable = (func(x, *args, **kwargs) for x in iterable)
    partial = functools.partial(func, *args, **kwargs)
    try:
        mapped_iterable = map(partial, iterable)
    except TypeError:
        mapped_iterable = (partial(x) async for x in iterable)
    async for task in limit_concurrency(mapped_iterable, limit = limit, return_when = return_when):
        yield await task


async def async_map(
    func: Callable[..., Awaitable[Any]],
    iterable: Iterable[Any], 
    *args,
    limit: Optional[int] = None,
    return_when: Optional[ReturnWhenType] = ReturnWhenType.FIRST_COMPLETED,
    **kwargs,
) -> AsyncGenerator[RT, None]:
    """
    Async Map of a function with args and kwargs

    Args:
        func (Callable[..., Awaitable[Any]]): The function to map
        iterable (Iterable[Any]): The iterable to map
        limit (Optional[int], optional): The limit of the concurrency. Defaults to None.
        return_when (Optional[ReturnWhenType], optional): The return when type. Defaults to ReturnWhenType.FIRST_COMPLETED.
    
    Yields:
        [type]: [description]
    """
    func = ensure_coro(func)
    partial = functools.partial(func, *args, **kwargs)
    try:
        mapped_iterable = map(partial, iterable)
    except TypeError:
        mapped_iterable = (partial(x) async for x in iterable)
    async for task in limit_concurrency(mapped_iterable, limit = limit, return_when = return_when):
        yield await task



from typing import Optional, Dict
from lazyops.types.models import BaseSettings

class ClientSettings(BaseSettings):

    """
    Client Settings
    """
    max_retries: Optional[int] = 5
    timeout: Optional[int] = 180
    debug_enabled: Optional[bool] = True

    api_key: Optional[str] = None
    api_header: Optional[str] = None

    # API Key Details
    api_dev_mode: Optional[bool] = False # sets to local_url if enabled

    local_url: Optional[str] = 'http://localhost:80'
    cluster_url: Optional[str] = 'http://api.svc.cluster.local:80'
    external_url: Optional[str] = None

    class Config:
        env_prefix = 'LAZYOPS_CLIENT_'
        case_sensitive = False
    
    def get_headers(self):
        _base_headers = {
            'content-type': 'application/json',
        }
        if self.api_key and self.api_header: _base_headers[self.api_header] = self.api_key
        return _base_headers
    
    @property
    def endpoints(self) -> Dict[str, str]:
        return {
            'local': self.local_url,
            'cluster': self.cluster_url,
            'external': self.external_url,
        }



"""
Common Cloud Provider Settings

pulled from `fileio.utils.configs`
"""
import os
import json
import pathlib

from typing import Optional, Union, Dict, Any
from lazyops.types.models import BaseSettings, validator
from lazyops.types.classprops import lazyproperty

from lazyops.imports._fileio import (
    File, 
    FileLike, 
    _fileio_available, 
    require_fileio,
)

# AWS
class BotoSettings(BaseSettings):
    boto_config: Optional[Union[str, pathlib.Path]] = None
    boto_path: Optional[Union[str, pathlib.Path]] = None

    @lazyproperty
    def path(self) -> FileLike:
        p = self.boto_config or self.boto_path
        if p is None: return None
        if _fileio_available: return File(p)
        if isinstance(p, str): 
            p = pathlib.Path(p)
        return p
    
    @lazyproperty
    def exists(self) -> bool:
        return False if self.path is None else self.path.exists()

    def set_env(self):
        """
        Update the Env variables for the current session
        """
        if self.exists:
            os.environ["BOTO_CONFIG"] = self.path.as_posix()
            os.environ["BOTO_PATH"] = self.path.as_posix()
    
class AwsSettings(BaseSettings):
    aws_access_token: Optional[str] = None
    aws_access_key_id: Optional[str] = None
    aws_secret_access_key: Optional[str] = None
    aws_region: Optional[str] = "us-east-1"
    set_s3_endpoint: Optional[bool] = True
    s3_config: Optional[Union[str, Dict[str, Any]]] = None
    s3_bucket: Optional[str] = None
    s3_backup_bucket: Optional[str] = None

    @validator("s3_config", pre=True)
    def validate_s3_config(cls, v):
        if v is None: return {}
        return json.loads(v) if isinstance(v, str) else v
    
    @lazyproperty
    def s3_endpoint(self):
        return f'https://s3.{self.aws_region}.amazonaws.com'
    
    @lazyproperty
    @require_fileio()
    def s3_bucket_path(self):
        if self.s3_bucket is None: return None
        bucket = self.s3_bucket
        if not bucket.startswith('s3://'): bucket = f's3://{bucket}'
        return File(bucket)
    
    @lazyproperty
    @require_fileio()
    def s3_backup_bucket_path(self):
        if self.s3_backup_bucket is None: return None
        bucket = self.s3_backup_bucket
        if not bucket.startswith('s3://'): bucket = f's3://{bucket}'
        return File(bucket)
    
    def set_env(self):
        if self.aws_access_key_id:
            os.environ['AWS_ACCESS_KEY_ID'] = self.aws_access_key_id
        if self.aws_secret_access_key:
            os.environ['AWS_SECRET_ACCESS_KEY'] = self.aws_secret_access_key
        if self.aws_region:
            os.environ['AWS_REGION'] = self.aws_region
        if self.aws_access_token:
            os.environ['AWS_ACCESS_TOKEN'] = self.aws_access_token
        if self.set_s3_endpoint:
            os.environ['S3_ENDPOINT'] = self.s3_endpoint
    


# GCP

class GcpSettings(BaseSettings):
    gcp_project: Optional[str] = None
    gcloud_project: Optional[str] = None
    google_cloud_project: Optional[str] = None
    google_application_credentials: Optional[Union[str, pathlib.Path]] = None

    gcs_client_config: Optional[Union[str, Dict[str, Any]]] = None
    gcs_config: Optional[Union[str, Dict[str, Any]]] = None

    gs_bucket: Optional[str] = None
    gs_backup_bucket: Optional[str] = None

    @validator("google_application_credentials")
    def validate_google_application_credentials(cls, v):
        if v is None: return pathlib.Path.home().joinpath('adc.json')
        if _fileio_available: return File(v)
        if isinstance(v, str): v = pathlib.Path(v)
        return v
    
    @validator("gcs_client_config")
    def validate_gcs_client_config(cls, v) -> Dict:
        if v is None: return {}
        return json.loads(v) if isinstance(v, str) else v
    
    @validator("gcs_config")
    def validate_gcs_config(cls, v) -> Dict:
        if v is None: return {}
        return json.loads(v) if isinstance(v, str) else v

    @lazyproperty
    def adc_exists(self):
        return self.google_application_credentials.exists()
    
    @lazyproperty
    def project(self):
        return self.gcp_project or self.gcloud_project or self.google_cloud_project
    
    @lazyproperty
    @require_fileio()
    def gs_bucket_path(self):
        if self.gs_bucket is None: return None
        bucket = self.gs_bucket
        if not bucket.startswith('gs://'): bucket = f'gs://{bucket}'
        return File(bucket)
    
    @lazyproperty
    @require_fileio()
    def gs_backup_bucket_path(self):
        if self.gs_backup_bucket is None: return None
        bucket = self.gs_backup_bucket
        if not bucket.startswith('gs://'): bucket = f'gs://{bucket}'
        return File(bucket)
    
    def set_env(self):
        if self.adc_exists:
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = self.google_application_credentials.as_posix()
        if self.project:
            os.environ["GOOGLE_CLOUD_PROJECT"] = self.project
    

# Minio

class MinioSettings(BaseSettings):
    minio_endpoint: Optional[str] = None
    minio_access_key: Optional[str] = None
    minio_secret_key: Optional[str] = None
    minio_access_token: Optional[str] = None
    minio_secure: Optional[bool] = True
    minio_region: Optional[str] = None
    minio_config: Optional[Union[str, Dict[str, Any]]] = None
    minio_signature_ver: Optional[str] = 's3v4'

    minio_bucket: Optional[str] = None
    minio_backup_bucket: Optional[str] = None

    @validator("minio_config", pre=True)
    def validate_minio_config(cls, v):
        if v is None: return {}
        return json.loads(v) if isinstance(v, str) else v
    
    @lazyproperty
    @require_fileio()
    def minio_bucket_path(self):
        if self.minio_bucket is None: return None
        bucket = self.minio_bucket
        if not bucket.startswith('minio://'): bucket = f'minio://{bucket}'
        return File(bucket)
    
    @lazyproperty
    @require_fileio()
    def minio_backup_bucket_path(self):
        if self.minio_backup_bucket is None: return None
        bucket = self.minio_backup_bucket
        if not bucket.startswith('minio://'): bucket = f'minio://{bucket}'
        return File(bucket)
    
    def set_env(self):
        if self.minio_endpoint:
            os.environ["MINIO_ENDPOINT"] = self.minio_endpoint
        if self.minio_access_key:
            os.environ["MINIO_ACCESS_KEY"] = self.minio_access_key
        if self.minio_secret_key:
            os.environ["MINIO_SECRET_KEY"] = self.minio_secret_key
        if self.minio_secure:
            os.environ["MINIO_SECURE"] = str(self.minio_secure)
        if self.minio_region:
            os.environ["MINIO_REGION"] = self.minio_region
        if self.minio_signature_ver:
            os.environ["MINIO_SIGNATURE_VER"] = self.minio_signature_ver











import json
import functools

from lazyops.types.models import BaseModel, validator
from lazyops.types.classprops import lazyproperty
from lazyops.types.static import RESPONSE_SUCCESS_CODES
from lazyops.types.resources import BaseResource, ResourceType, ResponseResource, ResponseResourceType
from lazyops.types.errors import ClientError, fatal_exception
from lazyops.imports._aiohttpx import aiohttpx, resolve_aiohttpx
from lazyops.imports._backoff import backoff, require_backoff
from lazyops.configs.base import DefaultSettings
from lazyops.utils.logs import default_logger as logger
from lazyops.utils.serialization import ObjectEncoder

from typing import Optional, Dict, List, Any, Type, Callable


"""
Client Session
"""
SettingType = Type[DefaultSettings]

_settings: SettingType = None

def get_settings(settings_cls: Optional[SettingType] = None, **kwargs) -> DefaultSettings:
    global _settings
    if _settings is None:
        if settings_cls is None: settings_cls = DefaultSettings
        _settings = settings_cls(**kwargs)
    return _settings


class ClientSession(BaseModel):
    # Base
    client: 'aiohttpx.Client'
    headers: Dict[str, str] = None
    success_codes: Optional[List[int]] = RESPONSE_SUCCESS_CODES

    # Options
    max_retries: Optional[int] = 3
    timeout: Optional[int] = None
    debug_enabled: Optional[bool] = False
    on_error: Optional[Callable] = None
    ignore_errors: Optional[bool] = False

    # Settings
    settings_cls: Optional[SettingType] = None
    settings: Optional[DefaultSettings] = None

    # Resource Models
    input_model: Optional[Type[ResourceType]] = BaseResource
    response_model: Optional[Type[BaseResource]] = ResponseResource
    error_model: Optional[Type[ClientError]] = ClientError

    @validator('settings_cls', pre = True)
    def validate_settings_cls(cls, v):
        if v is None: v = DefaultSettings
        return v

    @validator('settings', pre = True)
    def validate_settings(cls, v, values):
        if v is None: v = get_settings(settings_cls = values.get('settings_cls'))
        return v

    @validator('headers')
    def validate_headers(cls, v, values):
        if v is None: 
            try: v = values.get('settings').get_client_headers()
            except:  v = {}
        return v
    
    @validator('timeout')
    def validate_timeout(cls, v, values):
        if v is None: 
            try: v = values.get('settings').client.timeout
            except: v = None
        return v
    
    @validator('max_retries')
    def validate_max_retries(cls, v, values):
        if v is None: 
            try: v = values.get('settings').client.max_retries
            except: v = 3
        return v
    
    @validator('debug_enabled')
    def validate_debug_enabled(cls, v, values):
        if v is None: 
            try: v = values.get('settings').client.debug_enabled
            except: v = False
        return v

    @lazyproperty
    def api_endpoints(self) -> Dict[str, Dict[str, str]]:
        """
        Returns the api endpoints and methods
        """
        return {
            'get': {
                'url': '/v1/get',
                'method': 'GET',
            },
            'list': {
                'url': '/v1/list',
                'method': 'GET',
            },
            'create': {
                'url': '/v1/create',
                'method': 'POST',
            },
            'update': {
                'url': '/v1/update',
                'method': 'PUT',
            },
            'delete': {
                'url': '/v1/delete',
                'method': 'DELETE',
            },
        }

    @lazyproperty
    def expo_func(self):
        return functools.partial(backoff.expo, base = 4)
    
    @staticmethod
    def object_dumps(input_object: ResourceType) -> str:
        return json.dumps(input_object.dict(), cls = ObjectEncoder)
    
    def get(self,
        input_object: Optional[ResourceType] = None,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ResponseResourceType:
        """
        Get a resource
        """
        if input_object is None: 
            input_object, kwargs = self.input_model.parse_resource(**kwargs)
        data = self.object_dumps(input_object)
        api_method_data = self.api_endpoints['get']
        api_response = self._send(
            method = api_method_data['method'],
            url = api_method_data['url'],
            data = data,
            params = params,
            headers = self.headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(input_obj = input_object, response = data)
    
    async def async_get(self,
        input_object: Optional[ResourceType] = None,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ResponseResourceType:
        """
        Get a resource
        """
        if input_object is None: 
            input_object, kwargs = self.input_model.parse_resource(**kwargs)
        data = self.object_dumps(input_object)
        api_method_data = self.api_endpoints['get']
        api_response = await self._async_send(
            method = api_method_data['method'],
            url = api_method_data['url'],
            data = data,
            params = params,
            headers = self.headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(input_obj = input_object, response = data)
    
    def list(
        self,
        input_object: Optional[ResourceType] = None,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ResponseResourceType:
        """
        List Resources
        """
        if input_object is None: 
            input_object, kwargs = self.input_model.parse_resource(**kwargs)
        data = self.object_dumps(input_object)
        api_method_data = self.api_endpoints['list']
        api_response = self._send(
            method = api_method_data['method'],
            url = api_method_data['url'],
            data = data,
            params = params,
            headers = self.headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(input_obj = input_object, response = data)
    
    async def async_list(
        self,
        input_object: Optional[ResourceType] = None,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ResponseResourceType:
        """
        List Resources
        """
        if input_object is None: 
            input_object, kwargs = self.input_model.parse_resource(**kwargs)
        data = self.object_dumps(input_object)
        api_method_data = self.api_endpoints['list']
        api_response = await self._async_send(
            method = api_method_data['method'],
            url = api_method_data['url'],
            data = data,
            params = params,
            headers = self.headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(input_obj = input_object, response = data)
    
    def create(
        self,
        input_object: Optional[ResourceType] = None,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ResponseResourceType:
        """
        Create a resource
        """
        if input_object is None: 
            input_object, kwargs = self.input_model.parse_resource(**kwargs)
        data = self.object_dumps(input_object)
        api_method_data = self.api_endpoints['create']
        api_response = self._send(
            method = api_method_data['method'],
            url = api_method_data['url'],
            data = data,
            params = params,
            headers = self.headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(input_obj = input_object, response = data)
    
    async def async_create(
        self,
        input_object: Optional[ResourceType] = None,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ResponseResourceType:
        """
        Create a resource
        """
        if input_object is None: 
            input_object, kwargs = self.input_model.parse_resource(**kwargs)
        data = self.object_dumps(input_object)
        api_method_data = self.api_endpoints['create']
        api_response = await self._async_send(
            method = api_method_data['method'],
            url = api_method_data['url'],
            data = data,
            params = params,
            headers = self.headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(input_obj = input_object, response = data)
    
    def update(
        self,
        input_object: Optional[ResourceType] = None,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ResponseResourceType:
        """
        Update a resource
        """
        if input_object is None: 
            input_object, kwargs = self.input_model.parse_resource(**kwargs)
        data = self.object_dumps(input_object)
        api_method_data = self.api_endpoints['update']
        api_response = self._send(
            method = api_method_data['method'],
            url = api_method_data['url'],
            data = data,
            params = params,
            headers = self.headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(input_obj = input_object, response = data)
    
    async def async_update(
        self,
        input_object: Optional[ResourceType] = None,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ResponseResourceType:
        """
        Update a resource
        """
        if input_object is None: 
            input_object, kwargs = self.input_model.parse_resource(**kwargs)
        data = self.object_dumps(input_object)
        api_method_data = self.api_endpoints['update']
        api_response = await self._async_send(
            method = api_method_data['method'],
            url = api_method_data['url'],
            data = data,
            params = params,
            headers = self.headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(input_obj = input_object, response = data)
    
    def delete(
        self,
        input_object: Optional[ResourceType] = None,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ResponseResourceType:
        """
        Delete a resource
        """
        if input_object is None: 
            input_object, kwargs = self.input_model.parse_resource(**kwargs)
        data = self.object_dumps(input_object)
        api_method_data = self.api_endpoints['delete']
        api_response = self._send(
            method = api_method_data['method'],
            url = api_method_data['url'],
            data = data,
            params = params,
            headers = self.headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(input_obj = input_object, response = data)
    
    async def async_delete(
        self,
        input_object: Optional[ResourceType] = None,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ResponseResourceType:
        """
        Delete a resource
        """
        if input_object is None: 
            input_object, kwargs = self.input_model.parse_resource(**kwargs)
        data = self.object_dumps(input_object)
        api_method_data = self.api_endpoints['delete']
        api_response = await self._async_send(
            method = api_method_data['method'],
            url = api_method_data['url'],
            data = data,
            params = params,
            headers = self.headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(input_obj = input_object, response = data)

    def handle_response(
        self, 
        response: 'aiohttpx.Response',
        **kwargs
    ):
        """
        Handle the Response

        :param response: The Response
        """
        if self.debug_enabled:
            logger.info(f'[{response.status_code} - {response.request.url}] headers: {response.headers}, body: {response.text}')
        
        if response.status_code in self.success_codes:
            return response if response.text else None
        
        if self.ignore_errors: return None
        raise self.error_model(
            response = response
        )

    def prepare_response(
        self,
        input_obj: 'ResourceType',
        response: 'aiohttpx.Response',
        response_object: Optional[ResponseResourceType] = None,
        **kwargs
    ):
        """
        Prepare the Response Object
        
        :param data: The Response Data
        :param response_object: The Response Object
        """
        response_object = response_object or self.response_model
        if response_object:
            return response_object.parse_from(input_obj = input_obj, response = response, **kwargs)
        raise NotImplementedError('Response model not defined for this resource.')

    
    def _send(
        self,
        method: str,
        url: str,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
        timeout: Optional[int] = None,
        retries: Optional[int] = None,
        **kwargs
    ) -> 'aiohttpx.Response':
        if retries is None: retries = self.max_retries
        if timeout is None: timeout = self.timeout
        @backoff.on_exception(
            self.expo_func, Exception, max_tries = retries + 1, giveup = fatal_exception
        )
        def _retryable_send():
            return self.client.request(
                method = method,
                url = url,
                params = params,
                data = data,
                headers = headers,
                timeout = timeout,
                **kwargs
            )
        return _retryable_send()
    
    async def _async_send(
        self,
        method: str,
        url: str,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
        timeout: Optional[int] = None,
        retries: Optional[int] = None,
        **kwargs
    ) -> 'aiohttpx.Response':
        if retries is None: retries = self.max_retries
        if timeout is None: timeout = self.timeout
        @backoff.on_exception(
            self.expo_func, Exception, max_tries = retries + 1, giveup = fatal_exception
        )
        async def _retryable_async_send():
            return await self.client.async_request(
                method = method,
                url = url,
                params = params,
                data = data,
                headers = headers,
                timeout = timeout,
                **kwargs
            )
        return await _retryable_async_send()



"""
Global Client Class
"""

class BaseClient:
    """
    Base Client Class
    - intended to be subclassed.
    - manages the various sessions
    """

    ctx: Optional[ClientSession] = None
    current: Optional[str] = None
    sessions: Dict[str, ClientSession] = {}

    # Settings
    settings_cls: Optional[SettingType] = None
    settings: Optional[DefaultSettings] = None

    """
    Sessions
    """
    @classmethod
    def init_settings(
        cls,
        settings_cls: Optional[SettingType] = None,
        settings: Optional[DefaultSettings] = None,
        **kwargs
    ):
        """
        Initializes the settings
        """
        if settings_cls is not None:
            cls.settings_cls = settings_cls
            if cls.settings is None: cls.settings = cls.settings_cls()
        if settings is not None: cls.settings = settings
        if cls.settings is None and cls.settings_cls: cls.settings = cls.settings_cls()
        if cls.settings and kwargs: cls.settings.update_config(**kwargs)
    

    @classmethod
    def init_session(
        cls,
        name: str = 'default',
        set_current: bool = False,
        overwrite: Optional[bool] = None,
        **kwargs,
    ) -> ClientSession:
        """
        Initialize the session
        """
        cls.init_settings(**kwargs)
        if name == 'default': name = cls.settings.default_client_name
        if name in cls.sessions and overwrite is not True:
            logger.warning(f'Session {name} already exists')
            return
        resolve_aiohttpx(True)
        client = aiohttpx.Client(
            base_url = kwargs.get('base_url', cls.settings.client.endpoints[name]),
            timeout = cls.settings.client.timeout,
        )
        ctx = ClientSession(
            client = client,
            **kwargs
        )
        cls.sessions[name] = ctx
        logger.info(f'Initialized Session: {name}')
        if (set_current or overwrite) or cls.ctx is None:
            cls.ctx = ctx
            cls.current = name
            logger.info(f'Setting to Current Session: {name}')
        return ctx


    @classmethod
    def get_session(
        cls, 
        name: Optional[str] = None,
        **kwargs
    ) -> ClientSession:
        """
        Get a session by name

        :param name: The name of the session.
        :type name: str
        :return: A :class:`ClientSession` instance.
        """
        if not name: name = cls.current or cls.settings.default_client_name
        if name not in cls.sessions:
            cls.init_session(name = name, **kwargs)
        return cls.sessions[name]





from __future__ import annotations

"""
Usage:



"""

import re
import json
import copy
import contextlib
import operator

from abc import ABC
from pydantic import BaseModel
from typing import Any, Dict, List, Optional, Union, Callable, Tuple, TYPE_CHECKING
from lazyops.utils import logger
from lazyops.utils.lazy import lazy_import
from lazyops.libs.fastapi_utils.types.user_roles import UserRole

if TYPE_CHECKING:
    from fastapi import Request, FastAPI


KEY_START = '[x-repl-start]'
KEY_END = '[x-repl-end]'
KEY_SEP = '||'
DOMAIN_KEY = 'DOMAIN'

def json_replace(
    schema: Dict[str, Any],
    key: str,
    value: Optional[str] = None,

    verbose: Optional[bool] = False,
    # Control Vars
    key_start: Optional[str] = KEY_START,
    key_end: Optional[str] = KEY_END,
    sep_char: Optional[str] = KEY_SEP,
) -> Dict[str, Any]:
    """
    Replace the json
    """
    plaintext = json.dumps(schema, ensure_ascii=False)
    if f'{key_start} {key}' not in plaintext:
        return schema
    while f'{key_start} {key}' in plaintext:
        src_value = value
        start = plaintext.index(f'{key_start} {key}')
        end = plaintext.index(f'{key_end}', start)
        segment = plaintext[start:end + len(key_end)].strip()
        # Default Value
        if not src_value and sep_char in segment:
            src_value = segment.split(sep_char, 1)[-1].rsplit(' ', 1)[0].strip()
        plaintext = plaintext.replace(segment, src_value)
        if verbose: logger.info(f"Replaced `{segment}` -> `{src_value}`", colored = True, prefix = "|g|OpenAPI|e|")
        if f'{key_start} {key}' not in plaintext:
            break
    schema = json.loads(plaintext)
    return schema


"""
Dynamic Patching of OpenAPI Specs

- route_schemas: Dict[str, Dict[str, str]]
- excluded_schemas: List[str]
"""

_openapi_schemas: Dict[str, Dict[str, Any]] = {}

def patch_openapi_schema(
    openapi_schema: Dict[str, Any], # OpenAPI Schema

    schemas_patches: Dict[str, Dict[str, str]], # Schemas to patch
    excluded_schemas: List[str], # Schemas to exclude

    overwrite: Optional[bool] = None, 
    verbose: Optional[bool] = True, 
    module_name: Optional[str] = None, # this will get injected

    # Patches
    replace_patches: Optional[List[Tuple[str, Union[Callable, Optional[str]]]]] = None,
    replace_key_start: Optional[str] = KEY_START,
    replace_key_end: Optional[str] = KEY_END,
    replace_sep_char: Optional[str] = KEY_SEP,
    **kwargs
) -> Dict[str, Any]:
    """
    Patch the openapi schema
    """
    global _openapi_schemas
    if module_name not in _openapi_schemas or overwrite:
        if verbose: logger.info(f"[|g|{module_name}|e|] Patching OpenAPI Schema", colored = True)
        for path, patch in schemas_patches.items():
            with contextlib.suppress(Exception):
                patch_source = openapi_schema['components']['schemas'].pop(patch['source'], None)
                if patch_source is None: continue

                patch_source['title'] = patch['schema']
                openapi_schema['components']['schemas'][patch['schema']] = patch_source

                # Handle Paths
                openapi_schema['paths'][path]['post']['requestBody']['content']['multipart/form-data']['schema']['$ref'] = f'#/components/schemas/{patch["schema"]}'
        
        if not excluded_schemas: excluded_schemas = []
        for exclude in excluded_schemas:
            _ = openapi_schema['components']['schemas'].pop(exclude, None)
        
        if replace_patches:
            for patch in replace_patches:
                key, value = patch
                if callable(value): value = value()
                openapi_schema = json_replace(
                    schema = openapi_schema,
                    key = key,
                    value = value,
                    verbose = verbose,
                    key_start = replace_key_start,
                    key_end = replace_key_end,
                    sep_char = replace_sep_char,
                )

        if 'schemas' in openapi_schema.get('components', {}):
            openapi_schema['components']['schemas'] = dict(sorted(openapi_schema['components']['schemas'].items()))
        _openapi_schemas[module_name] = openapi_schema
    
    return _openapi_schemas[module_name]

def create_openapi_schema_patch(
    module_name: str,
    schemas_patches: Dict[str, Dict[str, str]],
    excluded_schemas: List[str],

    replace_patches: Optional[List[Tuple[str, Union[Callable, Optional[str]]]]] = None,
    replace_key_start: Optional[str] = KEY_START,
    replace_key_end: Optional[str] = KEY_END,
    replace_sep_char: Optional[str] = KEY_SEP,
    replace_domain_key: Optional[str] = DOMAIN_KEY,

) -> Callable[[Dict[str, Any]], Dict[str, Any]]:
    """
    Create an openapi schema patch
    """
    def patch_openapi_schema_wrapper(
        openapi_schema: Dict[str, Any],
        overwrite: Optional[bool] = None, 
        verbose: Optional[bool] = True, 
        **kwargs
    ) -> Dict[str, Any]:
        """
        Patch the openapi schema wrapper
        """
        return patch_openapi_schema(
            openapi_schema = openapi_schema,
            schemas_patches = schemas_patches,
            excluded_schemas = excluded_schemas,
            module_name = module_name,
            overwrite = overwrite,
            verbose = verbose,
            replace_patches = replace_patches,
            replace_key_start = replace_key_start,
            replace_key_end = replace_key_end,
            replace_sep_char = replace_sep_char,
        )
    return patch_openapi_schema_wrapper

def get_openapi_schema(module_name: str) -> Dict[str, Any]:
    """
    Get the openapi schema
    """
    global _openapi_schemas
    return _openapi_schemas[module_name]

_server_domains: Dict[str, str] = {}

def get_server_domain(
    request: Optional['Request'] = None,
    module_domains: Optional[List[str]] = None,
    module_name: Optional[str] = None, # this will get injected
    verbose: Optional[bool] = True,
    force_https: Optional[bool] = None,
) -> Optional[str]:
    """
    Get the server domain that the app is hosted on
    """
    global _server_domains
    if not _server_domains.get(module_name):
        if request is None: return None
        if not module_domains: 
            module_domains = ['localhost', module_name]
        if any(
            domain in request.url.hostname for 
            domain in module_domains
        ):
            scheme = 'https' if (force_https or request.url.port == 443) else request.url.scheme
            # _server_domains[module_name] = f'{request.url.scheme}://{request.url.hostname}'
            _server_domains[module_name] = f'{scheme}://{request.url.hostname}'
            if request.url.port and request.url.port not in {80, 443}:
                _server_domains[module_name] += f':{request.url.port}'
            if verbose: logger.info(f"[|g|{module_name}|e|] Setting Server Domain: {_server_domains[module_name]} from {request.url}", colored = True)
    return _server_domains.get(module_name)


"""
Dynamic Patching of OpenAPI Specs by Role
"""

class OpenAPIRoleSpec(ABC):

    role: Optional[Union['UserRole', str]] = None
    
    included_paths: Optional[List[Union[str, Dict[str, str]]]] = []
    excluded_paths: Optional[List[Union[str, Dict[str, str]]]] = []

    included_tags: Optional[List[str]] = []
    excluded_tags: Optional[List[str]] = []

    included_schemas: Optional[List[str]] = []
    excluded_schemas: Optional[List[str]] = []

    openapi_schema: Optional[Dict[str, Any]] = None
    description_callable: Optional[Callable] = None

    extra_schemas: Optional[List[Union[BaseModel, Dict[str, Any], str]]] = None

    def __init__(
        self, 
        role: Optional['UserRole'] = None,
        included_paths: Optional[List[str]] = None,
        excluded_paths: Optional[List[str]] = None,
        included_tags: Optional[List[str]] = None,
        excluded_tags: Optional[List[str]] = None,
        included_schemas: Optional[List[str]] = None,
        excluded_schemas: Optional[List[str]] = None,
        openapi_schema: Optional[Dict[str, Any]] = None,

        description_callable: Optional[Callable] = None,
        extra_schemas: Optional[List[Union[BaseModel, Dict[str, Any], str]]] = None,
        extra_schema_prefix: Optional[str] = None,
        extra_schema_name_mapping: Optional[Dict[str, str]] = None,
        extra_schema_ref_template: Optional[str] = '#/components/schemas/{model}',
        extra_schema_example_mapping: Optional[Dict[str, Dict[str, Any]]] = None,
        extra_schema_example_callable: Optional[Callable] = None,
        **kwargs
    ):
        self.role = role or UserRole.ANON
        self.included_paths = included_paths or []
        self.excluded_paths = excluded_paths or []
        self.included_tags = included_tags or []
        self.excluded_tags = excluded_tags or []
        self.included_schemas = included_schemas or []
        self.excluded_schemas = excluded_schemas or []
        self.openapi_schema = openapi_schema
        self.description_callable = description_callable
        if extra_schemas is not None:
            self.extra_schemas = extra_schemas
        self.extra_schema_prefix = extra_schema_prefix
        self.extra_schema_name_mapping = extra_schema_name_mapping
        self.extra_schemas_populated = False
        self.extra_schemas_data: Dict[str, Dict[str, Any]] = None
        self.extra_schema_ref_template = extra_schema_ref_template
        self.extra_schema_example_mapping = extra_schema_example_mapping
        self.extra_schema_example_callable = extra_schema_example_callable
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def has_description_callable(self) -> bool:
        """
        Check if the role spec has a description callable
        """
        return self.description_callable is not None
    
    def populate_extra_schemas(self):
        """
        Populate the extra schemas
        """
        if self.extra_schemas_populated: return
        if not self.extra_schemas: return
        self.extra_schemas_data = {}
        for schema in self.extra_schemas:
            if isinstance(schema, str):
                try:
                    schema = lazy_import(schema)
                except Exception as e:
                    logger.warning(f"Invalid Extra Schema: {schema}, {e}")
                    continue
            if isinstance(schema, type(BaseModel)):
                schema_name = schema.__name__
                try:
                    schema = schema.model_json_schema(
                        ref_template = self.extra_schema_ref_template
                    )
                except Exception as e:
                    logger.warning(f"Invalid Extra Schema: {schema}, {e}")
                    continue
            elif isinstance(schema, dict):
                if 'title' not in schema:
                    logger.warning(f"Invalid Extra Schema. Does not contain `title` in schema: {schema}")
                    continue
                schema_name = schema['title']
            else:
                logger.warning(f"Invalid Extra Schema: {schema}")
                continue
            if self.extra_schema_name_mapping and schema_name in self.extra_schema_name_mapping:
                schema['title'] = self.extra_schema_name_mapping[schema_name]
            elif self.extra_schema_prefix:
                schema['title'] = f'{self.extra_schema_prefix}{schema_name}'
            
            if self.extra_schema_example_callable:
                if schema_example := self.extra_schema_example_callable(schema = schema, schema_name = schema_name):
                    schema['example'] = schema_example
            elif self.extra_schema_example_mapping and schema_name in self.extra_schema_example_mapping:
                schema['example'] = self.extra_schema_example_mapping[schema_name]
            self.extra_schemas_data[schema_name] = schema
        self.extra_schemas_populated = True


    

_openapi_schemas_by_role: Dict[str, Dict[UserRole, OpenAPIRoleSpec]] = {}

def setup_new_module_schema(
    module_name: str,
    roles: List[OpenAPIRoleSpec],
):
    """
    Create a new module schema
    """
    global _openapi_schemas_by_role
    if module_name not in _openapi_schemas_by_role:
        _openapi_schemas_by_role[module_name] = {}
        for role in roles:
            _openapi_schemas_by_role[module_name][role.role] = role

def get_module_by_role_schema(module_name: str) -> Dict[UserRole, OpenAPIRoleSpec]:
    """
    Get the module by role schema
    """
    global _openapi_schemas_by_role
    return _openapi_schemas_by_role[module_name]

def extract_schemas_from_path_operation(
    spec: Dict[str, Dict[str, Any]],
) -> List[str]:
    """
    Extract the schemas from a path operation
    """
    schemas = []
    for response in spec['responses'].values():
        if 'content' not in response: continue
        for content in response['content'].values():
            if 'schema' not in content: continue
            if 'definitions' in content['schema']:
                schemas += content['schema']['definitions'].keys()
            if 'title' in content['schema']:
                schemas.append(content['schema']['title'])
            if '$ref' in content['schema']:
                schemas.append(content['schema']['$ref'].split('/')[-1])
    
    return schemas

def set_module_role_spec(
    module_name: str,
    role_spec: OpenAPIRoleSpec,
):
    """
    Set the module role spec
    """
    global _openapi_schemas_by_role
    _openapi_schemas_by_role[module_name][role_spec.role] = role_spec

def create_openapi_schema_by_role_function(
    module_name: str,
    roles: List[OpenAPIRoleSpec],
    module_domains: Optional[List[str]] = None,
    domain_name: Optional[str] = None,
    domain_name_force_https: Optional[bool] = None,

    default_exclude_paths: Optional[List[str]] = None,

    replace_patches: Optional[List[Tuple[str, Union[Callable, Optional[str]]]]] = None,
    replace_key_start: Optional[str] = KEY_START,
    replace_key_end: Optional[str] = KEY_END,
    replace_sep_char: Optional[str] = KEY_SEP,
    replace_domain_key: Optional[str] = DOMAIN_KEY,
    verbose: Optional[bool] = True,
    **kwargs
) -> Callable[[Dict[str, Any]], Dict[str, Any]]:
    """
    Create an openapi schema by role
    """

    setup_new_module_schema(module_name = module_name, roles = roles)
    if not default_exclude_paths: default_exclude_paths = []

    def patch_openapi_description(
        role_spec: OpenAPIRoleSpec,
        schema: Dict[str, Union[Dict[str, Union[Dict[str, Any], Any]], Any]],
        request: Optional['Request'] = None,
        force_https: Optional[bool] = None,
    ) -> Dict[str, Any]:
        """
        Patch the openapi schema description
        """
        nonlocal domain_name
        if domain_name is None: domain_name = get_server_domain(
            request = request, 
            module_name = module_name, 
            module_domains = module_domains, 
            verbose = verbose,
            force_https = domain_name_force_https or force_https,
        )
        if domain_name:
            replace_domain_start = f'<<{replace_domain_key}>>'
            replace_domain_end = f'>>{replace_domain_key}<<'
            schema['info']['description'] = schema['info']['description'].replace(
                replace_domain_start, domain_name
            ).replace(replace_domain_end, domain_name)
            if role_spec.has_description_callable:
                schema['info']['description'] = role_spec.description_callable(
                    schema['info']['description'],
                    domain_name,
                    role_spec,
                )
        return schema
    

    def patch_openapi_paths(
        role_spec: OpenAPIRoleSpec,
        schema: Dict[str, Union[Dict[str, Union[Dict[str, Any], Any]], Any]],
    ) -> Dict[str, Any]:
        """
        Patch the openapi schema paths
        """
        
        # Remove tags
        for path, methods in schema['paths'].items():
            for method, spec in methods.items():
                if 'tags' in spec and any(tag in role_spec.excluded_tags for tag in spec['tags']):
                    role_spec.excluded_paths.append(path)
                    # schemas_to_remove += await _extract_schemas_from_path_operation(spec)

        for path in role_spec.excluded_paths:
            if isinstance(path, str):
                schema['paths'].pop(path, None)
            elif isinstance(path, dict):
                schema['paths'][path['path']].pop(path['method'], None)
        
        for path in default_exclude_paths:
            if path in role_spec.included_paths: continue
            if isinstance(path, str):
                schema['paths'].pop(path, None)
            elif isinstance(path, dict):
                schema['paths'][path['path']].pop(path['method'], None)
        
        # Remove schemas
        # Compile the schemas to remove
        if 'components' not in schema: return schema
        if 'schemas' not in schema['components']: return schema
        _schemas_to_remove = []
        for schema_name in role_spec.excluded_schemas:
            if '*' in schema_name:
                _schemas_to_remove += [
                    schema for schema in schema['components']['schemas'].keys() if
                    re.match(schema_name, schema)
                ]
            else:
                _schemas_to_remove.append(schema_name)
        _schemas_to_remove = list(set(_schemas_to_remove))
        for schema_name in _schemas_to_remove:
            schema['components']['schemas'].pop(schema_name, None)
        if role_spec.extra_schemas:
            role_spec.populate_extra_schemas()
            # if verbose:
            #     logger.info(f"Populating Extra Schemas for {role_spec.role}\n{list(role_spec.extra_schemas_data.keys())}", prefix = module_name)
            schema['components']['schemas'].update(role_spec.extra_schemas_data)
        schema['components']['schemas'] = dict(sorted(schema['components']['schemas'].items(), key = lambda x: operator.itemgetter('title')(x[1])))
        return schema


    def get_openapi_schema_by_role(
        user_role: Optional[Union['UserRole', str]] = None,
        request: Optional['Request'] = None,
        app: Optional['FastAPI'] = None,
        force_https: Optional[bool] = None,
    ) -> Dict[str, Any]:
        """
        Get the openapi schema by role
        """

        module_schemas = get_module_by_role_schema(module_name = module_name)
        role = user_role or UserRole.ANON
        role_spec = module_schemas[role]
        if not role_spec.openapi_schema:
            if verbose: logger.info("Generating OpenAPI Schema", prefix = role)
            schema = copy.deepcopy(app.openapi()) if app else copy.deepcopy(get_openapi_schema(module_name))
            patch_openapi_description(
                role_spec = role_spec,
                schema = schema,
                request = request,
                force_https = domain_name_force_https or force_https,
            )
            patch_openapi_paths(
                role_spec = role_spec,
                schema = schema,
            )
            if replace_patches:
                for patch in replace_patches:
                    key, value = patch
                    if callable(value): value = value()
                    schema = json_replace(
                        schema = schema,
                        key = key,
                        value = value,
                        key_start = replace_key_start,
                        key_end = replace_key_end,
                        sep_char = replace_sep_char,
                    )
            role_spec.openapi_schema = schema
            set_module_role_spec(
                module_name = module_name,
                role_spec = role_spec,
            )
        return role_spec.openapi_schema
    
    return get_openapi_schema_by_role



import json
import typing
import datetime
import contextlib
from enum import Enum
from sqlalchemy import inspect
from lazyops.utils.serialization import object_serializer, Json
from sqlalchemy.ext.declarative import DeclarativeMeta
from pydantic import create_model, BaseModel, Field
from typing import Optional, Dict, Any, List, Union, Type, cast


# https://stackoverflow.com/questions/5022066/how-to-serialize-sqlalchemy-result-to-json

class AlchemyEncoder(json.JSONEncoder):

    def default(self, obj):
        if not isinstance(obj.__class__, DeclarativeMeta):
            return json.JSONEncoder.default(self, obj)
        # an SQLAlchemy class
        fields = {}
        for field in [x for x in dir(obj) if not x.startswith('_') and x != 'metadata']:
            data = obj.__getattribute__(field)
            try:
                json.dumps(data) # this will fail on non-encodable values, like other classes
                fields[field] = data
            except TypeError:
                fields[field] = None
        # a json-encodable dict
        return fields

class SQLJson(Json):

    @staticmethod
    def dumps(
        obj: typing.Dict[typing.Any, typing.Any], 
        *args, 
        default: typing.Dict[typing.Any, typing.Any] = None, 
        cls: typing.Type[json.JSONEncoder] = AlchemyEncoder,
        _fallback_method: typing.Optional[typing.Callable] = None,
        **kwargs
    ) -> str:
        try:
            return json.dumps(obj, *args, default = default, cls = cls, **kwargs)
        except Exception as e:
            if _fallback_method is not None:
                return _fallback_method(obj, *args, default = default, **kwargs)
            raise e


_pydantic_models: Dict[str, Type[BaseModel]] = {}
_json_encoders = {
    datetime.datetime: lambda v: v.isoformat(),
    Enum: lambda v: v.value
}
with contextlib.suppress(ImportError):
    import numpy as np
    _json_encoders[np.ndarray] = lambda v: v.tolist()
    _json_encoders[np.int64] = lambda v: int(v)
    _json_encoders[np.float64] = lambda v: float(v)


class BasePydanticConfig:
    orm_mode = True
    arbitrary_types_allowed = True
    json_encoders = _json_encoders
    json_loads = Json.loads
    json_dumps = Json.dumps
    extra = 'allow'

# from lazyops.utils import logger

def cast_to_optional(t: Type[Any]) -> Optional[Type[Any]]:
    """
    Cast a type to an optional type
    """
    # logger.warning(f'Cast to optional: {t}')
    # if isinstance(t, None):
    if t == type(None):
        # Turn it into Any
        # logger.warning(f'Cast to optional from none: {t}')
        return Optional[Any]
    if isinstance(t, type):
        return Optional[t]
    return Optional[t[0]] if isinstance(t, list) else Optional[t.__args__[0]]

def get_sqlmodel_dict(obj: DeclarativeMeta) -> Dict[str, Any]:
    """
    Return a dictionary representation of a sqlalchemy model
    """
    return {c.key: getattr(obj, c.key) for c in inspect(obj).mapper.column_attrs}

def get_model_fields(obj: DeclarativeMeta) -> Dict[str, Any]:
    """
    Return a dictionary representation of a sqlalchemy model
    """
    fields = {}
    for c in inspect(obj).mapper.column_attrs:
        # logger.warning(f'Cast to optional: {c.key}')
        fields[c.key] = (cast_to_optional(type(getattr(obj, c.key))), Field(None))
    # map relationships
    # for r in inspect(obj).mapper.relationships:
    # for r in obj.__mapper__.relationships:
    #     if r.mapper.class_ == obj.__class__: continue
    #     # determine if the relationship is a parent
    #     # if it is, then we don't need to create a model for it
    #     if r.direction.name == 'MANYTOONE':
    #         logger.warning(f'Skip Cast to optional: {r.key}: {r.uselist} {r.mapper.class_} {r.direction.name}')
    #         # fields[r.key] = (cast_to_optional(type(getattr(obj, r.key))), Field(None))
    #         continue
    #     logger.warning(f'Cast to optional: {r.key}: {r.uselist} {r.mapper.class_} {r.direction.name} {r}')
    #     base_type = build_pydantic_model(r.mapper.class_)
    #     if r.uselist:
    #         fields[r.key] = (cast_to_optional(List[base_type]), Field(None))
    #     else:
    #         fields[r.key] = (cast_to_optional(base_type), Field(None))
        # fields[r.key] = (cast_to_optional(List[get_pydantic_model(r.mapper.class_)]), Field(None))
    return fields

def build_pydantic_model(obj: DeclarativeMeta) -> Type[BaseModel]:
    """
    Create a pydantic model from a sqlalchemy model
    """
    global _pydantic_models
    
    obj_class_name = f'{obj.__class__.__module__}.{obj.__class__.__name__}Model'
    if obj_class_name not in _pydantic_models:
        # fields = {
        #     c.key: (cast_to_optional(type(getattr(obj, c.key))), Field(None))
        #     for c in inspect(obj).mapper.column_attrs
        # }
        fields = get_model_fields(obj)
        # logger.info(f'Creating pydantic model for {obj_class_name}: {fields}')
        _pydantic_models[obj_class_name] = create_model(
            f'{obj.__class__.__name__}Model',
            __config__ = BasePydanticConfig,
            __module__=obj.__class__.__module__,
            **fields,
        )
    return _pydantic_models[obj_class_name]


def get_pydantic_model(obj: object) -> Type[BaseModel]:
    """
    Create a pydantic model from a sqlalchemy model
    """
    # obj_class_name = f'{obj.__class__.__module__}.{obj.__class__.__name__}Model'
    return build_pydantic_model(obj).from_orm(obj)

    # global _pydantic_models
    
    # obj_class_name = f'{obj.__class__.__module__}.{obj.__class__.__name__}Model'
    # if obj_class_name not in _pydantic_models:
    #     # fields = {
    #     #     c.key: (cast_to_optional(type(getattr(obj, c.key))), Field(None))
    #     #     for c in inspect(obj).mapper.column_attrs
    #     # }
    #     fields = get_model_fields(obj)
    #     # logger.info(f'Creating pydantic model for {obj_class_name}: {fields}')
    #     _pydantic_models[obj_class_name] = create_model(
    #         f'{obj.__class__.__name__}Model',
    #         __config__ = BasePydanticConfig,
    #         __module__=obj.__class__.__module__,
    #         **fields,
    #     )
    #     # logger.info(f'Created pydantic model for {obj_class_name}: {_pydantic_models[obj_class_name]}: {fields}')
    # # return _pydantic_models[obj_class_name](**obj.dict())
    # # print(_pydantic_models[obj_class_name].__fields__)
    # return _pydantic_models[obj_class_name].from_orm(obj)


def dict_diff(dict_a: Union[Dict[str, Any], Type[BaseModel]], dict_b: Union[Dict[str, Any], Type[BaseModel]], show_value_diff: bool = True):
    
    """
    Compare two dictionaries and return the difference between them
    """
    if not isinstance(dict_a, dict) and hasattr(dict_a, 'dict'):
        dict_a = dict_a.dict()
    if not isinstance(dict_b, dict) and hasattr(dict_b, 'dict'):
        dict_b = dict_b.dict()
    result = {
        'added': {k: dict_b[k] for k in set(dict_b) - set(dict_a)},
        'removed': {k: dict_a[k] for k in set(dict_a) - set(dict_b)},
    }
    if show_value_diff:
        common_keys =  set(dict_a) & set(dict_b)
        result['value_diffs'] = {
            k:(dict_a[k], dict_b[k])
            for k in common_keys
            if dict_a[k] != dict_b[k]
        }
    return result